
from .build_item import BuildItem
from collections import OrderedDict as odict


# -----------------------------------------------------------------------------
class Variant(BuildItem):
    """for variations in build flags"""

    @staticmethod
    def default_str():
        return 'none'

    @staticmethod
    def default():
        return Variant(__class__.default_str())

    @staticmethod
    def create_variants(spec_list):
        d = odict([('variants', (Variant, spec_list))])
        d = BuildItem.create(d)
        assert d.collections == ['variants']
        return d['variants']



#!/usr/bin/env python3

import unittest as ut
import subtest_fix
import shlex
import shutil
import glob

import os.path as osp

from c4.cmany.main import cmany_main
from c4.cmany import util
from c4.cmany import cmake
from c4.cmany.compiler import Compiler
from c4.cmany.build import Build

mydir = osp.abspath(osp.dirname(__file__))
tc_armhf = osp.join(mydir, "toolchain-arm-linux-gnueabihf.cmake")
assert osp.exists(tc_armhf)
testdirs = [osp.join(mydir, "hello")]


def toolchain_compiler_exists(tc_file):
    comps = cmake.extract_toolchain_compilers(tc_file)
    return osp.exists(comps['CMAKE_CXX_COMPILER'])


def run_with_args(testdir, args_in):
    if isinstance(args_in, str):
        args_in = shlex.split(args_in)
    with util.setcwd(testdir):
        bdir = osp.join('.test', 'xcompile', 'build')
        idir = osp.join('.test', 'xcompile', 'install')
        if osp.exists(bdir):
            shutil.rmtree(bdir)
        if osp.exists(idir):
            shutil.rmtree(idir)
        #
        args = ['--show-args', 'build']
        args += ['--build-dir', bdir, '--install-dir', idir]
        args += args_in
        cmany_main(args)
        #
        with util.setcwd(bdir):
            actual_builds = sorted([str(s) for s in list(glob.glob('*'))])
        return actual_builds


def do_toolchain_builds(toolchain, test, args, expected_builds):
    if not toolchain_compiler_exists(toolchain):
        return
    args = [a.format(toolchain=toolchain) for a in args]
    comps = cmake.extract_toolchain_compilers(toolchain)
    c = Compiler(comps['CMAKE_CXX_COMPILER'])
    expected_builds = [b.format(compiler=Build.sanitize_compiler_name(c.name))
                       for b in expected_builds]
    for t in testdirs:
        with test.subTest(msg=osp.dirname(t), args=args):
            actual_builds = run_with_args(t, args)
            test.assertEqual(actual_builds, expected_builds)


class Test00Arm(ut.TestCase):

    def r(self, args, expected_builds):
        do_toolchain_builds(tc_armhf, self, args, expected_builds)

    def test00(self):
        self.r(["--toolchain", "{toolchain}",
                "-s", "linux",
                "-a", "armv5: -X '-march=armv5'",
                "-a", "armv7: -X '-march=armv7'",
        ], [
            'linux-armv5-{compiler}-Release',
            'linux-armv7-{compiler}-Release',
        ])

    def test01(self):
        self.r(["-s", "linux: --toolchain {toolchain}",
                "-a", "armv5: -X '-march=armv5'",
                "-a", "armv7: -X '-march=armv7'",
        ], [
            'linux-armv5-{compiler}-Release',
            'linux-armv7-{compiler}-Release',
        ])

    def test02(self):
        self.r(["-s", "linux: --toolchain {toolchain}",
                "-a", "armv5: -X '-march=armv5'",
                "-a", "armv7: -X '-march=armv7'",
        ], [
            'linux-armv5-{compiler}-Release',
            'linux-armv7-{compiler}-Release',
        ])

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    ut.main()



from .build_item import BuildItem


# -----------------------------------------------------------------------------
class BuildType(BuildItem):
    """Specifies a build type, ie, one of Release, Debug, etc"""

    @staticmethod
    def default_str():
        return "Release"

    @staticmethod
    def default():
        return BuildType(__class__.default_str())



#!/usr/bin/env python3

import os.path as osp
import sys
from collections import OrderedDict as odict

topics = odict()


class Topic:

    keylen = 0
    tablefmt = None

    def __init__(self, id, title, txt, disabled=False):
        self.id = id
        self.title = title
        self.__doc__ = title
        self.txt = txt
        Topic.keylen = max(Topic.keylen, len(id)) + 1
        Topic.tablefmt = '    {:' + str(Topic.keylen) + '} {}'
        if not disabled:
            topics[id] = self


def create_topic(id, title, txt, disabled=False):
    ht = Topic(id, title, txt, disabled)
    setattr(sys.modules[__name__], 'help_' + id, ht)


def _get_doc(topic_name):
    df = osp.abspath(osp.dirname(__file__))
    dd = osp.join(df, "doc")
    if osp.exists(dd):
        txtdocs = dd
    else:
        txtdocs = osp.abspath(osp.join(df, "../../../doc/_build/text"))
    if not osp.exists(txtdocs):
        raise Exception("df=" + df + " ..... txtdocs=" + txtdocs)
    with open(osp.join(txtdocs, topic_name + ".txt")) as f:
        txt = "".join(f.readlines())
    return txt


# -----------------------------------------------------------------------------
create_topic(
    "quick_tour",
    title="Quick tour",
    txt=_get_doc("quick_tour")
)


# -----------------------------------------------------------------------------
create_topic(
    "build_items",
    title="Specifying build items",
    txt=_get_doc("build_items")
)


# -----------------------------------------------------------------------------
create_topic(
    "flags",
    title="Specifying compiler flags and build-item specific properties",
    txt=_get_doc("flags")
)


# -----------------------------------------------------------------------------
create_topic(
    "vs",
    title="Specifying Microsoft Visual Studio versions and toolsets",
    txt=_get_doc("vs")
)


# -----------------------------------------------------------------------------
create_topic(
    "dependencies",
    title="Dealing with project dependencies",
    txt=_get_doc("dependencies")
)


# -----------------------------------------------------------------------------
create_topic(
    "reusing_arguments",
    title="Simplify the input to cmany.",
    txt=_get_doc("reusing_arguments")
)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
epilog = """

list of help topics:
{}
""".format(
    '\n'.join([Topic.tablefmt.format(k, v.title) for k, v in topics.items()])
)



import re


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class CombinationRule:
    """x=exclude / i=include"""

    def __init__(self, x_or_i, what, patterns):
        assert x_or_i in ('x', 'i')
        assert what in (
            'builds_any', 'builds_all',
            'systems', 'architectures', 'compilers', 'build_types', 'variants'
        )
        self.x_or_i = x_or_i
        self.what = what
        self.patterns = patterns

    def is_valid(self, s, a, c, t, v):
        s = s.name
        a = a.name
        c = c.name
        t = t.name
        v = v.name

        result = True
        if self.what in ('builds_any', 'builds_all'):
            matches_all = True
            matches_none = True
            matches_any = False
            for pattern in self.patterns:
                if __class__.pattern_matches(pattern, s, a, c, t, v):
                    matches_any = True
                    matches_none = False
                else:
                    matches_all = False
            if self.x_or_i == 'x':
                if matches_none:
                    result = True
                else:
                    if self.what == 'builds_any':
                        result = not matches_any
                    elif self.what == 'builds_all':
                        result = not matches_all
                    else:
                        raise Exception("never reach this point")
            elif self.x_or_i == 'i':
                if matches_none:
                    result = False
                else:
                    if self.what == 'builds_any':
                        result = matches_any
                    elif self.what == 'builds_all':
                        result = matches_all
                    else:
                        raise Exception("never reach this point")
            return result
        else:
            if self.what == 'systems':
                which = s
            elif self.what == 'architectures':
                which = a
            elif self.what == 'compilers':
                which = c
            elif self.what == 'build_types':
                which = t
            elif self.what == 'variants':
                which = v
            if not isinstance(which, str):
                which = which.name
            #
            in_patterns = which in self.patterns
            if self.x_or_i == 'x':
                result = not in_patterns
            elif self.x_or_i == 'i':
                result = in_patterns
            return result

    @staticmethod
    def pattern_matches(pattern, s, a, c, t, v):
        from .build import Build
        #for i in (s, a, c, t, v):
        #    if re.search(pattern, str(i)):
        #        # print("GOT A MATCH:", self.rule, i)
        #        return True
        s = Build.get_tag(s, a, c, t, v)
        # print(s, type(s))
        if re.search(pattern, s):
            # print("GOT A TAG MATCH:", self.rule, s)
            return True
        # print("NO MATCH:", self.rule, s)
        return False


# -----------------------------------------------------------------------------
class CombinationRules:

    def __init__(self, specs):
        self.rules = []
        for x_or_i, any_or_all, rules in specs:
            crc = CombinationRule(x_or_i, any_or_all, rules)
            self.rules.append(crc)

    def is_valid(self, s, a, c, t, v):
        result = 1
        for r in self.rules:
            result = result & r.is_valid(s, a, c, t, v)
        return result

    def valid_combinations(self, systems, archs, comps, types, variants):
        combs = []
        for s in systems:
            for a in archs:
                for c in comps:
                    for t in types:
                        for v in variants:
                            valid = True
                            if not self.is_valid(s, a, c, t, v):
                                valid = False
                            for item in (s, a, c, t, v):
                                crs = item.combination_rules
                                #for idx, r in enumerate(crs.rules):
                                #    print((s, a, c, t, v), ":", item, ": rule[{}]".format(idx), r.x_or_i, r.what, r.patterns)
                                if not crs.is_valid(s, a, c, t, v):
                                    valid = False
                                    break
                            if valid:
                                combs.append((s, a, c, t, v))
        return combs



#!/usr/bin/env python3

import unittest as ut
import subtest_fix
import argparse
from collections import OrderedDict as odict

import c4.cmany as cmany
from c4.cmany import util, args as c4args
from c4.cmany.project import CombinationRules

# utility variables for simpler names
ds = cmany.System.default_str()
da = cmany.Architecture.default_str()
dc = cmany.Compiler.default_str()
dt = cmany.BuildType.default_str()
dv = cmany.Variant.default_str()

vnone = 'none'
vfoo = 'foo: -X wall'
vbar = 'bar: -D BAR_DEF=1'
vspec = "'{}','{}','{}'".format(vnone, vfoo, vbar)


def do_combination_t(test, input, expected_items, expected_combinations):
    assert isinstance(expected_items, odict)

    _parser = argparse.ArgumentParser()
    c4args.add_select(_parser)
    c4args.add_bundle_flags(_parser)

    args = _parser.parse_args(input)
    # util.logcmd(input, "\n", args)

    for i in ('systems', 'architectures', 'compilers', 'build_types', 'variants'):
        expected = expected_items[i]
        expected = [e.split(':')[0] for e in expected]
        actual = getattr(args, i)
        actual = [a.split(':')[0] for a in actual]
        # print(i, "actual=", actual)
        # print(i, "expected=", expected)
        # with test.subTest(msg="expected count of "+i, input=input):
        #    test.assertEqual(len(actual), len(expected))
        with test.subTest(msg="expected list of "+i, input=input):
            test.assertEqual(actual, expected)

    # expected combinations (constructed with their types)
    expected_combinations = [(
        cmany.System(s),
        cmany.Architecture(a),
        cmany.Compiler(c),
        cmany.BuildType(t),
        cmany.Variant(v)
    ) for s, a, c, t, v in expected_combinations]

    # actual combinations
    s, a, c, t, v = cmany.Project.get_build_items(**vars(args))
    cr = []
    if hasattr(args, 'combination_rules'):
        cr = args.combination_rules
    cr = CombinationRules(cr)
    actual_combinations = cr.valid_combinations(s, a, c, t, v)

    # print("\n")
    # for i, (e, a) in enumerate(zip(expected_combinations, actual_combinations)):
    #     print("expected[{}]".format(i), e, "actual[{}]".format(i), a)
    with test.subTest(msg="expected count of combinations", input=input):
        test.assertEqual(len(expected_combinations), len(actual_combinations))
    for exp, act in zip(expected_combinations, actual_combinations):
        #print("exp x act:", exp, act)
        with test.subTest(msg="combination architecture", input=input):
            test.assertEqual(act[0].name, exp[0].name)
        with test.subTest(msg="combination system", input=input):
            test.assertEqual(act[1].name, exp[1].name)
        with test.subTest(msg="combination compiler", input=input):
            test.assertEqual(act[2].name, exp[2].name)
        with test.subTest(msg="combination build type", input=input):
            test.assertEqual(act[3].name, exp[3].name)
        with test.subTest(msg="combination variant", input=input):
            test.assertEqual(act[4].name, exp[4].name)


class CombinationTestCase(ut.TestCase):

    def t(self, input, expected_items, expected_combinations):
        do_combination_t(self, input, expected_items, expected_combinations)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test00BaseLine(CombinationTestCase):

    def test00(self):
        self.t(
            # args
            [],
            # expected items
            odict([
                ('systems', [ds]),
                ('architectures', [da]),
                ('compilers', [dc]),
                ('build_types', [dt]),
                ('variants', [dv])
            ]),
            # expected combinations
            [
                (ds, da, dc, dt, dv),
            ]
        )

    def test01(self):
        self.t(
            # args
            ["-a", "x86,x86_64", "-v", vspec],
            # expected items
            odict([
                ('systems', [ds]),
                ('architectures', ['x86', 'x86_64']),
                ('compilers', [dc]),
                ('build_types', [dt]),
                ('variants', [vnone, vfoo, vbar])
            ]),
            # expected combinations
            [
                (ds, 'x86', dc, dt, vnone),
                (ds, 'x86', dc, dt, vfoo),
                (ds, 'x86', dc, dt, vbar),
                (ds, 'x86_64', dc, dt, vnone),
                (ds, 'x86_64', dc, dt, vfoo),
                (ds, 'x86_64', dc, dt, vbar),
            ]
        )

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test01AtCommandLevel(CombinationTestCase):

    def test01(self):
        rules = "x86_64.*foo"
        args = ["-a", "x86,x86_64", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', [dt]),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            (ds, 'x86', dc, dt, vnone),
            (ds, 'x86', dc, dt, vfoo),
            (ds, 'x86', dc, dt, vbar),
            (ds, 'x86_64', dc, dt, vnone),
            # (ds, 'x86_64', dc, dt, vfoo),
            (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_inc = [
            # (ds, 'x86', dc, dt, vnone),
            # (ds, 'x86', dc, dt, vfoo),
            # (ds, 'x86', dc, dt, vbar),
            # (ds, 'x86_64', dc, dt, vnone),
            (ds, 'x86_64', dc, dt, vfoo),
            # (ds, 'x86_64', dc, dt, vbar),
        ]
        self.t(args + ["--exclude-builds", rules], expected_items,
               expected_combinations_exc)
        self.t(args + ["--exclude-builds-all", rules], expected_items,
               expected_combinations_exc)
        self.t(args + ["--include-builds", rules], expected_items,
               expected_combinations_inc)
        self.t(args + ["--include-builds-all", rules], expected_items,
               expected_combinations_inc)

    def test02(self):
        rules = "x86-.*foo"
        args = ["-a", "x86,x86_64", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', [dt]),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            (ds, 'x86', dc, dt, vnone),
            # (ds, 'x86', dc, dt, vfoo),
            (ds, 'x86', dc, dt, vbar),
            (ds, 'x86_64', dc, dt, vnone),
            (ds, 'x86_64', dc, dt, vfoo),
            (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_inc = [
            # (ds, 'x86', dc, dt, vnone),
            (ds, 'x86', dc, dt, vfoo),
            # (ds, 'x86', dc, dt, vbar),
            # (ds, 'x86_64', dc, dt, vnone),
            # (ds, 'x86_64', dc, dt, vfoo),
            # (ds, 'x86_64', dc, dt, vbar),
        ]
        self.t(args + ["--exclude-builds", rules], expected_items,
               expected_combinations_exc)
        self.t(args + ["--exclude-builds-all", rules], expected_items,
               expected_combinations_exc)
        self.t(args + ["--include-builds", rules], expected_items,
               expected_combinations_inc)
        self.t(args + ["--include-builds-all", rules], expected_items,
               expected_combinations_inc)

    def test03(self):
        rules = "x86-.*foo,x86_64-.*bar"
        args = ["-a", "x86,x86_64", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', [dt]),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc_any = [
            (ds, 'x86', dc, dt, vnone),
            # (ds, 'x86', dc, dt, vfoo),
            (ds, 'x86', dc, dt, vbar),
            (ds, 'x86_64', dc, dt, vnone),
            (ds, 'x86_64', dc, dt, vfoo),
            # (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_exc_all = [
            (ds, 'x86', dc, dt, vnone),
            (ds, 'x86', dc, dt, vfoo),
            (ds, 'x86', dc, dt, vbar),
            (ds, 'x86_64', dc, dt, vnone),
            (ds, 'x86_64', dc, dt, vfoo),
            (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_inc_any = [
            # (ds, 'x86', dc, dt, vnone),
            (ds, 'x86', dc, dt, vfoo),
            # (ds, 'x86', dc, dt, vbar),
            # (ds, 'x86_64', dc, dt, vnone),
            # (ds, 'x86_64', dc, dt, vfoo),
            (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_inc_all = [
            # (ds, 'x86', dc, dt, vnone),
            # (ds, 'x86', dc, dt, vfoo),
            # (ds, 'x86', dc, dt, vbar),
            # (ds, 'x86_64', dc, dt, vnone),
            # (ds, 'x86_64', dc, dt, vfoo),
            # (ds, 'x86_64', dc, dt, vbar),
        ]
        self.t(args + ["--exclude-builds", rules], expected_items,
               expected_combinations_exc_any)
        self.t(args + ["--exclude-builds-all", rules], expected_items,
               expected_combinations_exc_all)
        self.t(args + ["--include-builds", rules], expected_items,
               expected_combinations_inc_any)
        self.t(args + ["--include-builds-all", rules], expected_items,
               expected_combinations_inc_all)

    def test10exc_x86(self):
        args = ["-a", "x86,x86_64", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', [dt]),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            # (ds, 'x86', dc, dt, vnone),
            # (ds, 'x86', dc, dt, vfoo),
            # (ds, 'x86', dc, dt, vbar),
            (ds, 'x86_64', dc, dt, vnone),
            (ds, 'x86_64', dc, dt, vfoo),
            (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_inc = [
            (ds, 'x86', dc, dt, vnone),
            (ds, 'x86', dc, dt, vfoo),
            (ds, 'x86', dc, dt, vbar),
            # (ds, 'x86_64', dc, dt, vnone),
            # (ds, 'x86_64', dc, dt, vfoo),
            # (ds, 'x86_64', dc, dt, vbar),
        ]
        self.t(args + ["-xa", 'x86'], expected_items,
               expected_combinations_exc)
        self.t(args + ["--exclude-architectures", 'x86'], expected_items,
               expected_combinations_exc)
        self.t(args + ["-ia", 'x86'], expected_items,
               expected_combinations_inc)
        self.t(args + ["--include-architectures", 'x86'], expected_items,
               expected_combinations_inc)

    def test11exc_x86_64(self):
        args = ["-a", "x86,x86_64", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', [dt]),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            (ds, 'x86', dc, dt, vnone),
            (ds, 'x86', dc, dt, vfoo),
            (ds, 'x86', dc, dt, vbar),
            # (ds, 'x86_64', dc, dt, vnone),
            # (ds, 'x86_64', dc, dt, vfoo),
            # (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_inc = [
            # (ds, 'x86', dc, dt, vnone),
            # (ds, 'x86', dc, dt, vfoo),
            # (ds, 'x86', dc, dt, vbar),
            (ds, 'x86_64', dc, dt, vnone),
            (ds, 'x86_64', dc, dt, vfoo),
            (ds, 'x86_64', dc, dt, vbar),
        ]
        self.t(args + ["-xa", 'x86_64'], expected_items,
               expected_combinations_exc)
        self.t(args + ["--exclude-architectures", 'x86_64'], expected_items,
               expected_combinations_exc)
        self.t(args + ["-ia", 'x86_64'], expected_items,
               expected_combinations_inc)
        self.t(args + ["--include-architectures", 'x86_64'], expected_items,
               expected_combinations_inc)

    def test12exc_x86_and_x86_64(self):
        args = ["-a", "x86,x86_64", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', [dt]),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            # (ds, 'x86', dc, dt, vnone),
            # (ds, 'x86', dc, dt, vfoo),
            # (ds, 'x86', dc, dt, vbar),
            # (ds, 'x86_64', dc, dt, vnone),
            # (ds, 'x86_64', dc, dt, vfoo),
            # (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_inc = [
            (ds, 'x86', dc, dt, vnone),
            (ds, 'x86', dc, dt, vfoo),
            (ds, 'x86', dc, dt, vbar),
            (ds, 'x86_64', dc, dt, vnone),
            (ds, 'x86_64', dc, dt, vfoo),
            (ds, 'x86_64', dc, dt, vbar),
        ]
        self.t(args + ["-xa", 'x86_64,x86'], expected_items,
               expected_combinations_exc)
        self.t(args + ["--exclude-architectures", 'x86_64,x86'], expected_items,
               expected_combinations_exc)
        self.t(args + ["-ia", 'x86_64,x86'], expected_items,
               expected_combinations_inc)
        self.t(args + ["--include-architectures", 'x86_64,x86'], expected_items,
               expected_combinations_inc)

    def test13exc_other(self):
        args = ["-a", "x86,x86_64", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', [dt]),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            (ds, 'x86', dc, dt, vnone),
            (ds, 'x86', dc, dt, vfoo),
            (ds, 'x86', dc, dt, vbar),
            (ds, 'x86_64', dc, dt, vnone),
            (ds, 'x86_64', dc, dt, vfoo),
            (ds, 'x86_64', dc, dt, vbar),
        ]
        expected_combinations_inc = [
            # (ds, 'x86', dc, dt, vnone),
            # (ds, 'x86', dc, dt, vfoo),
            # (ds, 'x86', dc, dt, vbar),
            # (ds, 'x86_64', dc, dt, vnone),
            # (ds, 'x86_64', dc, dt, vfoo),
            # (ds, 'x86_64', dc, dt, vbar),
        ]
        self.t(args + ["-xa", 'other'], expected_items,
               expected_combinations_exc)
        self.t(args + ["--exclude-architectures", 'other'], expected_items,
               expected_combinations_exc)
        self.t(args + ["-ia", 'other'], expected_items,
               expected_combinations_inc)
        self.t(args + ["--include-architectures", 'other'], expected_items,
               expected_combinations_inc)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test02AtBuildItemLevel(CombinationTestCase):

    def test01(self):
        args_exc = ["-a", "'x86: -xv foo',x86_64", "-t", "Debug,Release", "-v", vspec]
        args_inc = ["-a", "'x86: -iv foo',x86_64", "-t", "Debug,Release", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', ['Debug', 'Release']),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            (ds, 'x86', dc, 'Debug', vnone),
            # (ds, 'x86', dc, 'Debug', vfoo),
            (ds, 'x86', dc, 'Debug', vbar),
            (ds, 'x86', dc, 'Release', vnone),
            # (ds, 'x86', dc, 'Release', vfoo),
            (ds, 'x86', dc, 'Release', vbar),
            (ds, 'x86_64', dc, 'Debug', vnone),
            (ds, 'x86_64', dc, 'Debug', vfoo),
            (ds, 'x86_64', dc, 'Debug', vbar),
            (ds, 'x86_64', dc, 'Release', vnone),
            (ds, 'x86_64', dc, 'Release', vfoo),
            (ds, 'x86_64', dc, 'Release', vbar),
        ]
        expected_combinations_inc = [
            #(ds, 'x86', dc, 'Debug', vnone),
            (ds, 'x86', dc, 'Debug', vfoo),
            #(ds, 'x86', dc, 'Debug', vbar),
            #(ds, 'x86', dc, 'Release', vnone),
            (ds, 'x86', dc, 'Release', vfoo),
            #(ds, 'x86', dc, 'Release', vbar),
            (ds, 'x86_64', dc, 'Debug', vnone),
            (ds, 'x86_64', dc, 'Debug', vfoo),
            (ds, 'x86_64', dc, 'Debug', vbar),
            (ds, 'x86_64', dc, 'Release', vnone),
            (ds, 'x86_64', dc, 'Release', vfoo),
            (ds, 'x86_64', dc, 'Release', vbar),
        ]
        self.t(args_exc, expected_items, expected_combinations_exc)
        self.t(args_inc, expected_items, expected_combinations_inc)

    def test02(self):
        args_exc = ["-a", "'x86: -xt Debug',x86_64", "-t", "Debug,Release", "-v", vspec]
        args_inc = ["-a", "'x86: -it Debug',x86_64", "-t", "Debug,Release", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', ['Debug', 'Release']),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            # (ds, 'x86', dc, 'Debug', vnone),
            # (ds, 'x86', dc, 'Debug', vfoo),
            # (ds, 'x86', dc, 'Debug', vbar),
            (ds, 'x86', dc, 'Release', vnone),
            (ds, 'x86', dc, 'Release', vfoo),
            (ds, 'x86', dc, 'Release', vbar),
            (ds, 'x86_64', dc, 'Debug', vnone),
            (ds, 'x86_64', dc, 'Debug', vfoo),
            (ds, 'x86_64', dc, 'Debug', vbar),
            (ds, 'x86_64', dc, 'Release', vnone),
            (ds, 'x86_64', dc, 'Release', vfoo),
            (ds, 'x86_64', dc, 'Release', vbar),
        ]
        expected_combinations_inc = [
            (ds, 'x86', dc, 'Debug', vnone),
            (ds, 'x86', dc, 'Debug', vfoo),
            (ds, 'x86', dc, 'Debug', vbar),
            #(ds, 'x86', dc, 'Release', vnone),
            #(ds, 'x86', dc, 'Release', vfoo),
            #(ds, 'x86', dc, 'Release', vbar),
            (ds, 'x86_64', dc, 'Debug', vnone),
            (ds, 'x86_64', dc, 'Debug', vfoo),
            (ds, 'x86_64', dc, 'Debug', vbar),
            (ds, 'x86_64', dc, 'Release', vnone),
            (ds, 'x86_64', dc, 'Release', vfoo),
            (ds, 'x86_64', dc, 'Release', vbar),
        ]
        self.t(args_exc, expected_items, expected_combinations_exc)
        self.t(args_inc, expected_items, expected_combinations_inc)

    def test03(self):
        args_exc = ["-a", "'x86: -xt Debug -xv foo',x86_64", "-t", "Debug,Release", "-v", vspec]
        args_inc = ["-a", "'x86: -it Debug -iv foo',x86_64", "-t", "Debug,Release", "-v", vspec]
        expected_items = odict([
            ('systems', [ds]),
            ('architectures', ['x86', 'x86_64']),
            ('compilers', [dc]),
            ('build_types', ['Debug', 'Release']),
            ('variants', [vnone, vfoo, vbar])
        ])
        expected_combinations_exc = [
            # (ds, 'x86', dc, 'Debug', vnone),
            # (ds, 'x86', dc, 'Debug', vfoo),
            # (ds, 'x86', dc, 'Debug', vbar),
            (ds, 'x86', dc, 'Release', vnone),
            #(ds, 'x86', dc, 'Release', vfoo),
            (ds, 'x86', dc, 'Release', vbar),
            (ds, 'x86_64', dc, 'Debug', vnone),
            (ds, 'x86_64', dc, 'Debug', vfoo),
            (ds, 'x86_64', dc, 'Debug', vbar),
            (ds, 'x86_64', dc, 'Release', vnone),
            (ds, 'x86_64', dc, 'Release', vfoo),
            (ds, 'x86_64', dc, 'Release', vbar),
        ]
        expected_combinations_inc = [
            #(ds, 'x86', dc, 'Debug', vnone),
            (ds, 'x86', dc, 'Debug', vfoo),
            #(ds, 'x86', dc, 'Debug', vbar),
            #(ds, 'x86', dc, 'Release', vnone),
            #(ds, 'x86', dc, 'Release', vfoo),
            #(ds, 'x86', dc, 'Release', vbar),
            (ds, 'x86_64', dc, 'Debug', vnone),
            (ds, 'x86_64', dc, 'Debug', vfoo),
            (ds, 'x86_64', dc, 'Debug', vbar),
            (ds, 'x86_64', dc, 'Release', vnone),
            (ds, 'x86_64', dc, 'Release', vfoo),
            (ds, 'x86_64', dc, 'Release', vbar),
        ]
        self.t(args_exc, expected_items, expected_combinations_exc)
        self.t(args_inc, expected_items, expected_combinations_inc)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    ut.main()



#!/usr/bin/env python3

import unittest as ut
import subtest_fix
import c4.cmany.util as util
import sys
import copy

from collections import OrderedDict as odict


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test00splitesc(ut.TestCase):
    def test(self):
        self.assertEqual(
            util.splitesc('hello\,world,yet\,another', ','),
            ['hello\,world', 'yet\,another'])


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test01splitesc_quoted(ut.TestCase):

    def t(self, input, expected, split_char=' '):
        with self.subTest(input=input):
            output = util.splitesc_quoted(input, split_char)
            self.assertEqual(output, expected)

    def _run(self, fn):
        fn('{} {}')
        fn(' {} {}')
        fn(' {} {} ')
        fn('{} {} ')
        fn('{}     {}')
        fn('     {}     {}')
        fn('     {}     {}     ')
        fn('{}     {}     ')

    def test00_spaces_only(self):
        t = lambda fmt: self.t(fmt, [])
        t(' ')
        t('    ')

    def test01_single_letters(self):
        lhs = 'a'
        rhs = 'b'
        self._run(lambda fmt: self.t(fmt.format(lhs, rhs), [lhs, rhs]))

    def test02_multi_letters(self):
        lhs = 'a c d e f g h'
        rhs = 'b c d e f g h'
        res = lhs.split(' ') + rhs.split(' ')
        self._run(lambda fmt: self.t(fmt.format(lhs, rhs), res))

    def test03_multi_letters_quote_at_begin_not_at_end(self):
        lhs = '"a c d e f g h'
        rhs = 'b c d e f g h'
        res = lhs.split(' ') + rhs.split(' ')
        self._run(lambda fmt: self.t(fmt.format(lhs, rhs), res))

    def test04_multi_letters_quote_at_end_not_at_begin(self):
        lhs = 'a c d e f g h'
        rhs = 'b c d e f g h"'
        res = lhs.split(' ') + rhs.split(' ')
        self._run(lambda fmt: self.t(fmt.format(lhs, rhs), res))


    def test10_multi_letters_quoted(self):
        lhs = '"a c d e f g h"'
        rhs = '"b c d e f g h"'
        self._run(lambda fmt: self.t(fmt.format(lhs, rhs), [lhs, rhs]))
        self.t('"a b c" "d e f" "g h i"', ['"a b c"', '"d e f"', '"g h i"'], ' ')
        self.t('"a,b,c","d,e,f","g,h,i"', ['"a,b,c"', '"d,e,f"', '"g,h,i"'], ',')
        self.t('abc,def,ghi', ['abc', 'def', 'ghi'], ',')
        self.t('abc, def, ghi', ['abc', ' def', ' ghi'], ',')
        self.t('abc ,def ,ghi', ['abc ', 'def ', 'ghi'], ',')
        self.t('abc , def , ghi', ['abc ', ' def ', ' ghi'], ',')

    def test11_multi_letters_quoted_escaped(self):
        lhs = '"a\\" \\"c\\" \\"d\\" \\"e\\" \\"f\\" \\"g\\" \\"h"'
        rhs = '"b\\" \\"c\\" \\"d\\" \\"e\\" \\"f\\" \\"g\\" \\"h"'
        self._run(lambda fmt: self.t(fmt.format(lhs, rhs), [lhs, rhs]))
        self.t('abc\\ def\\ ghi', ['abc\\ def\\ ghi'], ' ')
        self.t('abc\\,def\\,ghi', ['abc\\,def\\,ghi'], ',')
        self.t('abc\\ def\\ ghi abc\\ def\\ ghi', ['abc\\ def\\ ghi', 'abc\\ def\\ ghi'], ' ')
        self.t('abc\\,def\\,ghi,abc\\,def\\,ghi', ['abc\\,def\\,ghi', 'abc\\,def\\,ghi'], ',')


    def test30_varspecs0(self):
        self.t('-X "-fPIC" -D VARIANT1', ['-X', '"-fPIC"', '-D', 'VARIANT1'])
        self.t('-X "-Wall" -D VARIANT2', ['-X', '"-Wall"', '-D', 'VARIANT2'])
        self.t('-X nortti,c++14 -D VARIANT3', ['-X', 'nortti,c++14', '-D', 'VARIANT3'])

    def test31_varspecs1(self):
        self.t('-X "-fPIC" -D VARIANT1,VARIANT_TYPE=1', ['-X', '"-fPIC"', '-D', 'VARIANT1,VARIANT_TYPE=1'])
        self.t('-X "-Wall" -D VARIANT2,VARIANT_TYPE=2', ['-X', '"-Wall"', '-D', 'VARIANT2,VARIANT_TYPE=2'])
        self.t('-X nortti,c++14 -D VARIANT3,VARIANT_TYPE=3', ['-X', 'nortti,c++14', '-D', 'VARIANT3,VARIANT_TYPE=3'])

    def test32_varspecs2(self):
        self.t('-X "-fPIC" -D VARIANT1,"VARIANT_TYPE=1"', ['-X', '"-fPIC"', '-D', 'VARIANT1,"VARIANT_TYPE=1"'])
        self.t('-X "-Wall" -D VARIANT2,"VARIANT_TYPE=2"', ['-X', '"-Wall"', '-D', 'VARIANT2,"VARIANT_TYPE=2"'])
        self.t('-X nortti,c++14 -D VARIANT3,"VARIANT_TYPE=3"', ['-X', 'nortti,c++14', '-D', 'VARIANT3,"VARIANT_TYPE=3"'])

    def test33_varspecs3(self):
        self.t('-X "-fPIC" -D "VARIANT1,VARIANT_TYPE=1"', ['-X', '"-fPIC"', '-D', '"VARIANT1,VARIANT_TYPE=1"'])
        self.t('-X "-Wall" -D "VARIANT2,VARIANT_TYPE=2"', ['-X', '"-Wall"', '-D', '"VARIANT2,VARIANT_TYPE=2"'])
        self.t('-X nortti,c++14 -D "VARIANT3,VARIANT_TYPE=3"', ['-X', 'nortti,c++14', '-D', '"VARIANT3,VARIANT_TYPE=3"'])


    def test40_multiple_varspecs0(self):
        var0 = '"var0: -X \\"-g\\",\\"-g3\\",\\"-Wall\\" -D VAR0,DBG"'
        var1 = '"var1: -X \\"-O1\\",\\"-ffast-math\\" -D VAR1,OPTM"'
        self.t('{},{}'.format(var0, var1), [var0, var1], ',')


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test02is_quoted(ut.TestCase):

    def test00simple(self):
        t = lambda s, expected: self.assertEqual(util.is_quoted(s), expected, msg=s)
        t('aasdasdasdasdasd', False)
        t('aasdasda""dasdasd', False)
        t('aasdasda''dasdasd', False)
        t('"aasdasdasdasdasd"', True)
        t("'aasdasdasdasdasd'", True)
        t('"aasdasda,sdasdasd"', True)
        t("'aasdasda,sdasdasd'", True)
        t('"aasdasda","sdasdasd"', False)
        t("'aasdasda','sdasdasd'", False)
        t('"aasdasda","sdasdasd","sdasdasd","sdasdasd"', False)
        t("'aasdasda','sdasdasd','aasdasda','sdasdasd'", False)
        t("'aasdasda',sdasdasd,aasdasda,'sdasdasd'", False)

    def test01nested(self):
        t = lambda s, expected: self.assertEqual(util.is_quoted(s), expected, msg=s)
        t("\'aasdasda,sdasdasd\'", True)
        t("\\'aasdasda,sdasdasd\\'", False)
        t('\'aasdasda,sdasdasd\'', True)
        t('\\\'aasdasda,sdasdasd\\\'', False)
        t("'aasdasda',\"sdasdasd\",'aasdasda',\"sdasdasd\"", False)
        t("'aasdasda',\'sdasdasd\'", False)
        #
        t("\'aasdasda,sdasdasd\'", True)
        t("\\'aasdasda,sdasdasd\\'", False)
        #
        t("'\'aasdasda\',\'sdasdasd\''", False)
        t("'\\'aasdasda\\',\\'sdasdasd\\''", True)

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test04has_interior_quotes(ut.TestCase):

    def test(self):
        t = lambda s, expected: self.assertEqual(util.has_interior_quotes(s), expected, msg=s)
        t('aaaaaa,bbbbbbb', False)
        t('"aaaaaa,bbbbbb"', False)
        t('aaaaaa,"bbbbbb"', True)
        t('"aaaaaa",bbbbbb', True)
        t('aaaaaa",bbbbbb', True)
        t('aaaaaa,"bbbbbb', True)
        t("none,'xdebug: --cxxflags g3'", True)
        t('none,"xdebug: --cxxflags g3"', True)
        t("none,\"xdebug: --cxxflags g3\"", True)
        t('none,\'xdebug: --cxxflags g3\'', True)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test05intersperse(ut.TestCase):

    def test_left(self):
        def t(li):
            expected = []
            for l in li:
                expected += ['delim', l]
            output = list(util.intersperse_l('delim', li))
            with self.subTest(input=li, version='left'):
                self.assertEqual(output, expected)
        t([0, 1, 2])
        t([2, 1, 0])


    def test_right(self):
        def t(li):
            expected = []
            for l in li:
                expected += [l, 'delim']
            output = list(util.intersperse_r('delim', li))
            with self.subTest(input=li, version='right'):
                self.assertEqual(output, expected)

        t([0, 1, 2])
        t([2, 1, 0])


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test09nested_merge(ut.TestCase):

    def test00reqs(self):
        import collections
        self.assertTrue(isinstance(dict(), collections.Mapping))
        self.assertTrue(isinstance(collections.OrderedDict(), collections.Mapping))

    def test01do_it(self):
        for fname in ('tc1', 'tc2', 'tc3'):
            fn = getattr(__class__, fname)
            for cls in (dict, odict):
                name, a, b, rab, rba = fn(self, cls)
                with self.subTest(cls=cls.__name__, fn=fname, case=name):
                    vab = util.nested_merge(a, b)
                    self.assertEqual(rab, vab)
                    vab = copy.deepcopy(a)
                    util.nested_merge(vab, b, False)
                    self.assertEqual(rab, vab)
                    #
                    vba = util.nested_merge(b, a)
                    self.assertEqual(rba, vba)
                    vba = copy.deepcopy(b)
                    util.nested_merge(vba, a, False)
                    self.assertEqual(rba, vba)

    def tc1(self, cls):
        return 'empty+empty=empty', cls(), cls(), cls(), cls()
    def tc2(self, cls):
        a = self.someval(cls)
        b = cls()
        return 'flat+empty=flat', a, b, a, a
    def tc3(self, cls):
        a = self.nested(cls, 'c')
        b = cls()
        return 'nested_after+empty=nested_after', a, b, a, a
    def tc4(self, cls):
        a = self.nested(cls, '0')
        b = cls()
        return 'nested_before+empty=nested_before', a, b, a, a

    def flat(self, cls):
        return self.someval(cls)
    def nested(self, cls, char):
        a = self.someval(cls)
        a[char] = self.someval(cls)
        a[char][char] = self.someval(cls)
        a[char][char][char] = self.someval(cls)
        return a
    def someval(self, cls):
        d = cls()
        d['a'] = 0
        d['b'] = 1
        return d


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test10runsyscmd(ut.TestCase):

    def test00_noargs(self):
        invoke_and_compare(self, [])

    def test01_noargs(self):
        invoke_and_compare(self, ['arg1', 'arg2'])

    def test02_quoted_args(self):
        invoke_and_compare(self, ['"arg1"', '"arg2"'])

    def test03_quoted_args_with_spaces(self):
        invoke_and_compare(self, ['"arg1 and more"', '"arg2 and more"'])


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
echo_args = None
def invoke_and_compare(tester, cmd_args):
    # print("invoke_and_compare: in=", cmd_args)
    cmd = [sys.executable, sys.modules[__name__].__file__, 'echo'] + cmd_args
    out = util.runsyscmd(cmd, echo_cmd=False, capture_output=True).strip()
    # print("invoke_and_compare: out=", out)
    # this doesn't work...:
    # code = "echo_args={}".format(out)
    # ... but this does:
    code = "tmp = " + out + "; setattr(sys.modules[__name__], 'echo_args', tmp)"
    code = code.format(out)
    # print("invoke_and_compare: code=", code)
    exec(code)
    # print("invoke_and_compare: echo_args=", echo_args)
    tester.assertEqual(echo_args, cmd_args)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == 'echo':
        print(sys.argv[2:])  # for test_invoke()
    else:
        ut.main()



#!/usr/bin/env python3

import unittest as ut
import subtest_fix

import os.path as osp
import tempfile

from c4.cmany.conf import Configs
from c4.cmany import conf
from c4.cmany.build_flags import BuildFlags

from ruamel import yaml as yaml
from ruamel.yaml.comments import CommentedMap as CommentedMap

# -----------------------------------------------------------------------------
def tmpfile(contents):
    fn = tempfile.mkstemp('.yml', 'cmany.')
    with open(fn, "w"):
        fn.write(contents)
    return fn


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test00ConfigsClass(ut.TestCase):

    def test00Empty(self):
        c = Configs()
        self.assertIsNotNone(c.get_val('project'))
        self.assertIsNotNone(c.get_val('config'))
        self.assertIsNotNone(c.get_val('flag_aliases'))

    def test01Load(self):
        c = Configs()
        c.load(text="foo: bar")
        self.assertIsNotNone(c.get_val('project'))
        self.assertIsNotNone(c.get_val('config'))
        self.assertIsNotNone(c.get_val('flag_aliases'))
        self.assertEqual(c.get_val('foo'), 'bar')

    def test02Merge(self):
        c1 = Configs()
        c1.load(text="""\
foo1: bar1
foo2: bar1
foo3: bar1
fonix:
  a: 1
  b: 1
  c: 1
  not_yet_another:
    a: 1
    b: 1
    c: 1
""")
        c2 = Configs()
        c2.load(text="""\
foo1: bar2
foo2: bar2
foo4: bar2
fonix:
  a: 2
  d: 2
  e: 2
  f: 2
  yet_another:
    a: 2
    d: 2
    e: 2
    f: 2
""")
        c1.merge_from(c2)
        self.assertEqual(c1.get_val('foo1'), 'bar2')
        self.assertEqual(c1.get_val('foo2'), 'bar2')
        self.assertEqual(c1.get_val('foo3'), 'bar1')
        self.assertEqual(c1.get_val('foo4'), 'bar2')
        self.assertEqual(c1.get_val('fonix.a'), 2)
        self.assertEqual(c1.get_val('fonix.b'), 1)
        self.assertEqual(c1.get_val('fonix.c'), 1)
        self.assertEqual(c1.get_val('fonix.d'), 2)
        self.assertEqual(c1.get_val('fonix.e'), 2)
        self.assertEqual(c1.get_val('fonix.f'), 2)
        self.assertEqual(c1.get_val('fonix.not_yet_another.a'), 1)
        self.assertEqual(c1.get_val('fonix.not_yet_another.b'), 1)
        self.assertEqual(c1.get_val('fonix.not_yet_another.c'), 1)
        self.assertEqual(c1.get_val('fonix.yet_another.a'), 2)
        self.assertEqual(c1.get_val('fonix.yet_another.d'), 2)
        self.assertEqual(c1.get_val('fonix.yet_another.e'), 2)
        self.assertEqual(c1.get_val('fonix.yet_another.f'), 2)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test01ConfigsFile(ut.TestCase):

    def test01ConfDirExists(self):
        self.assertTrue(osp.exists(conf.CONF_DIR))

    def test01ConfFileExists(self):
        fn = osp.join(conf.CONF_DIR, "cmany.yml")
        self.assertTrue(osp.exists(fn))

    def test02CanLoadConfFile(self):
        fn = osp.join(conf.CONF_DIR, "cmany.yml")
        c = Configs.load_seq((fn,))
        self.assertEqual(c.get_val('project'), {})

    def test03Set(self):
        fn = osp.join(conf.CONF_DIR, "cmany.yml")
        c = Configs.load_seq((fn,))
        self.assertIsNone(c.get_val("foo.bar.baz"))
        c.set_val("foo.bar.baz", 10)
        c.set_val("foo.bar.askdjaksjd", 11)
        self.assertEqual(c.get_val("foo.bar.baz"), 10)
        self.assertEqual(c.get_val("foo.bar.askdjaksjd"), 11)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class ConfFixtureBase(ut.TestCase):
    """base class for config test cases"""

    def do_load(self):
        cm = self.cm.copy()
        bf = BuildFlags("")
        bf.load_config(cm)
        return bf

    def do_save(self):
        cm = CommentedMap()
        self.bf.save_config(cm)
        return cm

    def do_roundtrip_save_load(self):
        cm = CommentedMap()
        self.bf.save_config(cm)
        bf = BuildFlags("")
        bf.load_config(cm)
        return bf

    def do_roundtrip_load_save(self):
        bf = BuildFlags("")
        bf.load_config(self.cm)
        cm = CommentedMap()
        bf.save_config(cm)
        return cm


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# Test cases for build flags

# using the _ prefix prevents unittest from instantiating tests for this class
class _BFTCase(ConfFixtureBase):
    """base for build flags test case"""

    def _setUp(self, bf, yml):
        self.bf = bf
        self.yml = yml
        self.name = self.__class__.__name__[7:].lower()
        YAML = yaml.YAML()
        self.cm = YAML.load(self.yml)
        if self.cm is None:
            self.cm = CommentedMap()
        self.non_empty = []
        for a in BuildFlags.attrs:
            if getattr(bf, a):
                self.non_empty.append(a)

    def check_bf(self, bf):
        with self.subTest(name=self.name, msg=self.name):
            for a in self.non_empty:
                self.assertEqual(getattr(bf, a), getattr(self.bf, a))
            for a in BuildFlags.attrs:
                if a not in self.non_empty:
                    self.assertTrue(not getattr(bf, a))

    def check_cm(self, cm):
        with self.subTest(name=self.name, msg=self.name):
            for a in self.non_empty:
                v = cm[a]
                if a != 'toolchain':
                    v = BuildFlags.flag_str_to_list(v)
                self.assertEqual(v, getattr(self.bf, a))
            for a in BuildFlags.attrs:
                if a not in self.non_empty:
                    self.assertTrue(not hasattr(cm, a))

    def test01Save(self):
        cm = self.do_save()
        self.check_cm(cm)

    def test02Load(self):
        bf = self.do_load()
        self.check_bf(bf)

    def test03RoundtripSaveLoad(self):
        bf = self.do_roundtrip_save_load()
        self.check_bf(bf)

    def test04RoundtripLoadSave(self):
        cm = self.do_roundtrip_load_save()
        self.check_cm(cm)


# -----------------------------------------------------------------------------
class Test02BuildFlags01Empty(_BFTCase):
    def setUp(self):
        self._setUp(bf=BuildFlags("empty"), yml="")


class Test02BuildFlags02CMakeVars(_BFTCase):
    def setUp(self):
        super()._setUp(
            bf=BuildFlags("cmake_vars",
                          cmake_vars=["VARFOO=BAR", "CMAKE_VERBOSE_MAKEFILE=1"]),
            yml="""\
cmake_vars: VARFOO=BAR CMAKE_VERBOSE_MAKEFILE=1
""")


class Test02BuildFlags03CMakeVarsAndCxxflags(_BFTCase):
    def setUp(self):
        self._setUp(
            bf=BuildFlags("cmake_vars_and_cxxflags",
                          cmake_vars=["VARFOO=BAR", "CMAKE_VERBOSE_MAKEFILE=1"],
                          cxxflags=["-std=c++11", "-Wall", ]),
            yml="""\
cmake_vars: VARFOO=BAR CMAKE_VERBOSE_MAKEFILE=1
cxxflags: -std=c++11 -Wall
""")


class Test02BuildFlags04Full(_BFTCase):
    def setUp(self):
        self._setUp(
            bf=BuildFlags("cmake_vars_and_cxxflags",
                          cmake_vars=["VARFOO=BAR", "CMAKE_VERBOSE_MAKEFILE=1"],
                          defines=["FOO", "BAR", "VFOO=baaaaaaaaaa"],
                          cflags=["-std=c++11", "-Wall", ],
                          cxxflags=["-std=c++11", "-Wall", ],
                          toolchain="somefile.cmake"),
            yml="""\
cmake_vars: VARFOO=BAR CMAKE_VERBOSE_MAKEFILE=1
defines: FOO BAR VFOO=baaaaaaaaaa
cflags: -std=c++11 -Wall
cxxflags: -std=c++11 -Wall
toolchain: somefile.cmake
""")


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    ut.main()



from .build_item import BuildItem
from .cmake import CMakeSysInfo


# -----------------------------------------------------------------------------
class System(BuildItem):
    """Specifies an operating system"""

    @staticmethod
    def default():
        """return the current operating system"""
        return System(__class__.default_str())

    @staticmethod
    def default_str():
        s = CMakeSysInfo.system_name()
        if s == "mac os x" or s == "Darwin":
            s = "mac"
        return s



#!/usr/bin/env python3

import argparse
import pprint
import os

from . import util
from . import help
from . import system
from . import build_item
from . import architecture
from . import compiler

from .util import cslist
from multiprocessing import cpu_count as cpu_count


# -----------------------------------------------------------------------------
def setup(subcommands, module):
    """creates a parser to process cmany's arguments"""
    p = argparse.ArgumentParser(
        prog='cmany',
        description='''Easily process several build trees of a CMake project''',
        usage='%(prog)s',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=help.epilog
    )
    sp = p.add_subparsers(help='')
    add_hidden(p)
    for cmd, aliases in subcommands.items():
        cl = getattr(module, cmd)
        h = sp.add_parser(name=cmd, aliases=aliases, help=cl.__doc__)
        cl().add_args(h)
        def exec_cmd(args, cmd_class=cl):
            obj = cmd_class()
            proj = obj.proj(args)
            obj._exec(proj, args)
        h.set_defaults(func=exec_cmd)
    return p


def parse(parser, in_args):
    """parses and performs related tasks"""
    args = parser.parse_args(in_args)
    if not hasattr(args, 'func'):
        argerror(parser, 'missing subcommand')
    if _handle_hidden_args__skip_rest(args):
        return None
    return args


def argerror(parser, *msg_args):
    print(*msg_args, end='')
    print('\n')
    parse(parser, ['-h'])
    exit(1)


def merge_envargs(cmds, sysargs):
    """gets any arguments from environment vars"""
    pfxargs = os.environ.get('CMANY_PFX_ARGS', '')
    pfxargs = util.splitesc_quoted(pfxargs, ' ')
    cmdargs = os.environ.get('CMANY_ARGS', '')
    cmdargs = util.splitesc_quoted(cmdargs, ' ')
    if not pfxargs and not cmdargs:
        return sysargs
    pos = find_subcommand(cmds, sysargs)
    args = sysargs
    cmd = sysargs[pos]
    if cmd not in ('help', 'h'):
        args = sysargs[0:pos]
        if pfxargs:
            # print("inserting CMANY_PFX_ARGS:", pfxargs)
            args += pfxargs
        args.append(cmd)
        if cmdargs:
            # print("inserting CMANY_ARGS:", cmdargs)
            args += cmdargs
        if len(sysargs) > pos+1:
            args += sysargs[(pos+1):]
    # print("resulting args:", args)
    return args


def find_subcommand(cmds, args):
    pos = None
    for i, a in enumerate(args):
        for c, aliases in cmds.items():
            if a == c or a in aliases:
                pos = i
                return pos
    raise Exception("could not find subcommand")


# -----------------------------------------------------------------------------
def add_hidden(parser):
    parser.add_argument('--show-args', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--only-show-args', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--show-args-list', type=cslist, default=[], help=argparse.SUPPRESS)
    parser.add_argument('--only-show-args-list', type=cslist, default=[], help=argparse.SUPPRESS)


def _handle_hidden_args__skip_rest(args):
    if args.show_args or args.only_show_args:
        pprint.pprint(vars(args), indent=4)
        if args.only_show_args:
            return True
    if args.show_args_list or args.only_show_args_list:
        l = args.show_args_list + args.only_show_args_list
        for a in l:
            print("args[", a, "]: ", sep='', end='')
            if not hasattr(args, a):
                print("(does not exist!)")
                continue
            v = getattr(args, a)
            pprint.pprint(v, indent=4)
        if args.only_show_args_list:
            return True
    return False


# -----------------------------------------------------------------------------
def add_proj(parser):
    parser.add_argument("proj_dir", nargs="?", default=".",
                        help="""the directory where the project's CMakeLists.txt
                        is located. An empty argument will default to the
                        current directory ie, \".\". Passing a directory which
                        does not contain a CMakeLists.txt will cause an error.""")
    parser.add_argument("--build-dir", default="./build",
                        help="set the build root (defaults to ./build)")
    parser.add_argument("--install-dir", default="./install",
                        help="set the install root (defaults to ./install)")
    parser.add_argument("-j", "--jobs", default=cpu_count(),
                        help="""build with the given number of parallel jobs
                        (defaults to %(default)s on this machine).""")
    parser.add_argument("-E", "--export-compile", default=False,
                        action="store_true",
                        help="""Have cmake export a compile_commands.json
                        containing the compile commands for each file. This
                        is useful eg for clang-based indexing tools.""")

    g = parser.add_argument_group('Configuration files')
    g.add_argument("--config-file", default=[], action="append",
                   help="""Specify a file containing configurations. Relative
                   paths are taken from the project's CMakeLists.txt directory.
                   Run `cmany help flags` to get help about flag aliases.
                   Multiple invokations are possible, in which case flags
                   given in latter files will prevail over those of earlier
                   files.""")
    g.add_argument("--no-default-config", default=False, action="store_true",
                   help="""Do not read the default config file. Run
                   `cmany help flags` to get help about this.""")

    d = parser.add_argument_group('Dependencies')
    d.add_argument('--deps', default='', type=str,
                   metavar='path/to/extern/CMakeLists.txt',
                   help="""Before configuring, process (ie, configure, build
                   and install) the given CMakeLists.txt project containing
                   needed external project dependencies. This will be done
                   separately for each build, using the same parameters. The
                   main project will be configured such that the built
                   dependencies are found by cmake.""")
    d.add_argument('--deps-prefix', default="", type=str,
                   metavar='path/to/install/directory',
                   help="""When using --deps set the install directory for
                   external dependencies to the given dir.""")
    d.add_argument('--with-conan', action='store_true', default=False,
                   help="""(WIP)""")


# -----------------------------------------------------------------------------
def _item_printer(dft):
    s = "["
    for i, item in enumerate(dft):
        if i > 1: s += ","
        s += str(item).replace(' ', '\ ')
    s += "]"
    return s


def add_select(parser):
    g = parser.add_argument_group(
        title="Build items",
        description="""Items to be combined by cmany. Each item can be made to
        bring specific flags with it by using the syntax 'item_name:
        <flags>...' (the quotes are needed when this syntax is used). Run
        `cmany help flags` to see the possible flags. Items can be
        given either as a comma-separated list or with repeated invokations
        of their arguments. Commas can be escaped by using a backslash,
        \\.""")

    dft = [system.System.default_str()]
    g.add_argument("-s", "--systems", metavar="os1,os2,...",
                   default=dft, action=BuildItemArgument,
                   help="""Specify a comma-separated list of operating systems
                   to combine. Defaults to the current system, """ +
                   _item_printer(dft) + """.""")

    dft = [architecture.Architecture.default_str()]
    g.add_argument("-a", "--architectures", metavar="arch1,arch2,...",
                   default=dft, action=BuildItemArgument,
                   help="""Specify a comma-separated list of processor
                   architectures to combine. Defaults to CMake's default
                   architecture on this system, """ +
                   _item_printer(dft) + """.""")

    dft = [compiler.Compiler.default_str()]
    g.add_argument("-c", "--compilers", metavar="compiler1,compiler2,...",
                   default=dft, action=BuildItemArgument,
                   help="""Specify a comma-separated list of compilers to
                   combine. Compilers can be given as an absolute path, or as
                   a name, in which case that name will be searched for in
                   the current shell's PATH.  Defaults to CMake's default
                   compiler on this system, """ +
                   _item_printer(dft) + """.""")

    # dft = [build_type.BuildType.default_str()]  # avoid a circular dependency
    dft = ["Release"]
    g.add_argument("-t", "--build-types", metavar="type1,type2,...",
                   default=dft, action=BuildItemArgument,
                   help="""Specify a comma-separated list of build types
                   to combine. Defaults to """ + _item_printer(dft) + """.""")

    # dft = [variant.Variant.default_str()]  # avoid a circular dependency
    dft = ["none"]
    g.add_argument("-v", "--variants", metavar="variant1,variant2,...",
                   default=["none"], action=BuildItemArgument,
                   help="""Specify a comma-separated list of variants
                   to combine. The variant name 'none' is special and will be
                   omitted from the name of the resulting build. Defaults to
                   """ + _item_printer(dft) + """.""")
    #add_combination_flags(parser)


# -----------------------------------------------------------------------------
def add_bundle_flags(parser):
    add_cflags(parser)
    add_combination_flags(parser)
    add_item_combination_flags(parser)


def add_combination_flags(parser):
    g = parser.add_argument_group(
        'Combination rules',
        description="""Prevent certain build item combinations from producing
                    a build. The rules in each argument are Python regular
                    expressions that are matched against each build name.
                    A build is included only if its name successfully matches
                    every combination argument. The build name is of the form
                    {system}-{architecture}-{compiler}-{build_type}[-{variant}].
                    The arguments are matched in the given order.""")
    g.add_argument("-xb", "--exclude-builds", metavar="rule1,[rule2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Exclude build item combinations whose name
                   matches ANY rule in the list.""")
    g.add_argument("-ib", "--include-builds", metavar="rule1,[rule2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Only allow build item combinations whose name
                   matches ANY rules in the list.""")
    g.add_argument("-xba", "--exclude-builds-all", metavar="rule1,[rule2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Exclude build item combinations whose name
                   matches ALL rules in the list.""")
    g.add_argument("-iba", "--include-builds-all", metavar="rule1,[rule2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Only allow build item combinations whose name
                   matches ALL rules in the list.""")


def add_item_combination_flags(parser):
    g = parser.add_argument_group(
        'Build-item-specific combination rules',
        description="""Prevent certain build items from producing a build.
        Each item is given by name and must be valid. These arguments are
        generally for use in specifications of build items; however, they may
        also be of use when using arguments via the CMANY_ARGS environment
        or with a project file.""")

    g.add_argument("-xs", "--exclude-systems", metavar="sys1,[sys2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Exclude builds which combine this item with any of the
                   following systems.""")
    g.add_argument("-is", "--include-systems", metavar="sys1,[sys2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Include only builds which combine this item with any
                   of the following systems.""")

    g.add_argument("-xa", "--exclude-architectures", metavar="arch1,[arch2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Exclude builds which combine this item with any of the
                   following architectures.""")
    g.add_argument("-ia", "--include-architectures", metavar="arch1,[arch2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Include only builds which combine this item with any of the
                   following architectures.""")

    g.add_argument("-xc", "--exclude-compilers", metavar="comp1,[comp2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Exclude builds which combine this item with any of the
                   following compilers.""")
    g.add_argument("-ic", "--include-compilers", metavar="comp1,[comp2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Include only builds which combine this item with any of the
                   following compilers.""")

    g.add_argument("-xt", "--exclude-build-types", metavar="btype1,[btype2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Exclude builds which combine this item with any of the
                   following build types.""")
    g.add_argument("-it", "--include-build-types", metavar="btype1,[btype2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Exclude only builds which combine this item with any of the
                   following build types.""")

    g.add_argument("-xv", "--exclude-variants", metavar="var1,[var2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Exclude builds which combine this item with any of the
                   following variants.""")
    g.add_argument("-iv", "--include-variants", metavar="var1,[var2[,...]]",
                   default=[], action=CombinationArgument,
                   help="""Include only builds which combine this item with any of the
                   following variants.""")


def add_cflags(parser):
    g = parser.add_argument_group(
        'Flags: CMake cache variables, compiler flags and defines',
        description="""Can be given both at command-level and at item-level.
        When given at command level, these arguments apply to all builds.
        When given at item-level (eg compilers or build types), these arguments
        apply only to the particular item.""")
    g.add_argument("-T", "--toolchain", metavar='toolchain_file', type=str, default=None,
                   help="""Specify a cmake toolchain file.""")
    g.add_argument("-V", "--cmake-vars", default=[], action=FlagArgument,
                   help="""Add CMake cache variables.
                   Accepts a comma-separated list of CMake cache
                   variable definitions.
                   Multiple invokations of -V are possible, in which case
                   arguments will be appended and not overwritten.
                   To escape commas, use a backslash \\.""")
    g.add_argument("-D", "--defines", default=[], action=FlagArgument,
                   help="""Add preprocessor symbol definitions.
                   Accepts a comma-separated list of preprocessor macro
                   definitions, in either of the forms -D FOO or -D FOO=value.
                   Multiple invokations of -D are possible, in which case
                   arguments will be appended and not overwritten.
                   To escape commas, use a backslash \\.""")
    g.add_argument("-X", "--cxxflags", default=[], action=FlagArgument,
                   help="""Add C++ compiler flags.
                   Accepts a comma-separated list of C++ compiler flags.
                   These will be passed to cmake by appending to the
                   default initial value of CMAKE_CXX_FLAGS (taken from
                   CMAKE_CXX_FLAGS_INIT). cmany has flag aliases mapping
                   a single key to different compilers. Run `cmany help flags`
                   to get help about this.
                   Multiple invokations of -X are possible, in which case
                   arguments will be appended and not overwritten.
                   To escape commas, use a backslash \\.""")
    g.add_argument("-C", "--cflags", default=[], action=FlagArgument,
                   help="""Add C compiler flags.
                   Accepts a comma-separated list of C compiler flags.
                   These will be passed to cmake by appending to the
                   default initial value of CMAKE_C_FLAGS (taken from
                   CMAKE_C_FLAGS_INIT). cmany has flag aliases mapping
                   a single key to different compilers. Run `cmany help flags`
                   to get help about this.
                   Multiple invokations of -X are possible, in which case
                   arguments will be appended and not overwritten.
                   To escape commas, use a backslash \\.""")
    # g.add_argument("-I", "--include-dirs", default=[], action=FlagArgument,
    #                help="""add dirs to the include path of all builds
    #                Multiple invokations of -I are possible, in which case arguments will be appended and not overwritten.
    #                Can also be given as a comma-separated list. To escape commas, use a backslash \\.""")
    # g.add_argument("-L", "--link-dirs", default=[], action=FlagArgument,
    #                help="""add dirs to the link path of all builds
    #                Multiple invokations of -L are possible, in which case arguments will be appended and not overwritten.
    #                Can also be given as a comma-separated list. To escape commas, use a backslash \\.""")


# -----------------------------------------------------------------------------
class FlagArgument(argparse.Action):
    """a class to be used by argparse when parsing arguments which
    are compiler flags"""
    def __call__(self, parser, namespace, values, option_string=None):
        li = getattr(namespace, self.dest)
        v = values
        # remove start and end quotes if there are any
        if util.is_quoted(v):
            v = util.unquote(v)
        # split at commas
        livals = cslist(v)
        # unquote splits
        for l in livals:
            ul = util.unquote(l)
            li.append(ul)
        setattr(namespace, self.dest, li)


# -----------------------------------------------------------------------------
class BuildItemArgument(argparse.Action):
    """a class to be used by argparse when parsing build item specifications.
    Note that prettyprint shows the specs wrong.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        li = getattr(namespace, self.dest)
        vli = build_item.BuildItem.parse_args(values)
        # clear the defaults from the list
        if not hasattr(parser, 'non_default_args'):
            parser.non_default_args = {}
        if parser.non_default_args.get(self.dest) is None:
            # util.logwarn("parsing: reset current li:", li)
            li = []
            parser.non_default_args[self.dest] = True
        # util.logwarn("parsing: current li:", li, " + ", vli)
        li += vli
        # util.logwarn("parsing: resulting li:", li)
        setattr(namespace, self.dest, li)


# -----------------------------------------------------------------------------
class CombinationArgument(argparse.Action):
    """parse combination arguments in order, putting them all into a single
    entry (combination_rules), which retains the original order. Maybe there's
    a cleverer way to do this but for now this is fast to implement."""
    def __call__(self, parser, namespace, values, option_string=None):
        # util.logwarn("parsing combinations: receive", self.dest, values)
        li = util.splitesc_quoted(values, ',')
        li = [util.unquote(item) for item in li]
        prev = getattr(namespace, self.dest)
        # util.logwarn("parsing combinations: receive", self.dest, values, ".... li", prev, "---->", prev + li)
        setattr(namespace, self.dest, prev + li)
        if not hasattr(namespace, 'combination_rules'):
            setattr(namespace, 'combination_rules', [])
        prev = getattr(namespace, 'combination_rules')

        if self.dest == 'exclude_builds':
            li = ('x', 'builds_any', li)
        elif self.dest == 'include_builds':
            li = ('i', 'builds_any', li)
        elif self.dest == 'exclude_builds_all':
            li = ('x', 'builds_all', li)
        elif self.dest == 'include_builds_all':
            li = ('i', 'builds_all', li)

        elif self.dest == 'exclude_systems':
            li = ('x', 'systems', li)
        elif self.dest == 'include_systems':
            li = ('i', 'systems', li)

        elif self.dest == 'exclude_architectures':
            li = ('x', 'architectures', li)
        elif self.dest == 'include_architectures':
            li = ('i', 'architectures', li)

        elif self.dest == 'exclude_compilers':
            li = ('x', 'compilers', li)
        elif self.dest == 'include_compilers':
            li = ('i', 'compilers', li)

        elif self.dest == 'exclude_build_types':
            li = ('x', 'build_types', li)
        elif self.dest == 'include_build_types':
            li = ('i', 'build_types', li)

        elif self.dest == 'exclude_variants':
            li = ('x', 'variants', li)
        elif self.dest == 'include_variants':
            li = ('i', 'variants', li)
        else:
            raise Exception("unoknown argument: " + self.dest)
        curr = prev
        curr.append(li)
        # util.logwarn("parsing combinations: receive", self.dest, values, ".... li", prev, "---->", curr)
        setattr(namespace, 'combination_rules', curr)



#!/usr/bin/env python3

import sys
import argcomplete

from collections import OrderedDict as odict

from c4.cmany.project import Project as Project
from c4.cmany import args as c4args
from c4.cmany import help as c4help


cmds = odict([
    ('help', ['h']),
    ('configure', ['c']),
    ('build', ['b']),
    ('install', ['i']),
    ('run', ['r']),
    ('show_vars', ['sv']),
    ('show_builds', ['sb']),
    ('show_build_names', ['sn']),
    ('show_build_dirs', ['sd']),
    ('show_targets', ['st']),
    ('create_proj', ['cp']),
    ('export_vs', []),
])


def cmany_main(in_args=None):
    if in_args is None:
        in_args = c4args.merge_envargs(cmds, sys.argv[1:])
    mymod = sys.modules[__name__]
    parser = c4args.setup(cmds, mymod)
    argcomplete.autocomplete(parser)
    args = c4args.parse(parser, in_args)
    if args:
        args.func(args)


# -----------------------------------------------------------------------------
class cmdbase:
    """base class for commands"""
    def add_args(self, parser):
        """add arguments to a command parser"""
        pass
    def proj(self, args):
        return None
    def _exec(self, proj, args):
        assert False, 'never call the base class method. Implement this in derived classes'


class help(cmdbase):
    """get help on a particular subcommand or topic"""
    def add_args(self, parser):
        super().add_args(parser)
        parser.add_argument('subcommand_or_topic', default="", nargs='?')
    def _exec(self, proj, args):
        sct = args.subcommand_or_topic
        if not sct:
            cmany_main(['-h'])
        else:
            sc = cmds.get(sct)
            # is it a subcommand?
            if sc is not None:
                cmany_main([sct, '-h'])
            else:
                # is it a topic?
                subtopic = c4help.topics.get(args.subcommand_or_topic)
                if subtopic is None:
                    msg = ("{} is not a subcommand or topic.\n" +
                           "Available subcommands are: {}\n" +
                           "Available help topics are: {}\n")
                    print(msg.format(args.subcommand_or_topic,
                                     ', '.join(cmds.keys()),
                                     ', '.join(c4help.topics.keys())))
                    exit(1)
                else:
                    print(subtopic.txt)


# -----------------------------------------------------------------------------
class projcmd(cmdbase):
    """a command which refers to a project"""
    def proj(self, args):
        """create a project given the configuration."""
        return Project(**vars(args))
    def add_args(self, parser):
        super().add_args(parser)
        c4args.add_proj(parser)
        c4args.add_bundle_flags(parser)


class selectcmd(projcmd):
    """a command which selects several builds"""
    def add_args(self, parser):
        super().add_args(parser)
        c4args.add_select(parser)


class configure(selectcmd):
    """configure the selected builds"""
    def _exec(self, proj, args):
        proj.configure()
    def add_args(self, parser):
        super().add_args(parser)


class build(selectcmd):
    """build the selected builds, configuring before if necessary"""
    def add_args(self, parser):
        super().add_args(parser)
        parser.add_argument('target', default=[], nargs='*',
                            help="""specify a subset of targets to build""")
    def _exec(self, proj, args):
        proj.build()


class install(selectcmd):
    """install the selected builds, building before if necessary"""
    def _exec(self, proj, args):
        proj.install()


class run(selectcmd):
    """run a command in each build directory"""
    def add_args(self, parser):
        super().add_args(parser)
        parser.add_argument('command', default="",
                            help="""command to be run in each build directory""")
    def _exec(self, proj, args):
        proj.run_cmd(args.command)


class show_vars(selectcmd):
    """show the value of certain CMake cache vars"""
    def add_args(self, parser):
        super().add_args(parser)
        parser.add_argument('var_names', default="", nargs='+')
    def _exec(self, proj, args):
        proj.show_vars(args.var_names)


class show_build_names(selectcmd):
    """show the build names"""
    def _exec(self, proj, args):
        proj.show_build_names()


class show_build_dirs(selectcmd):
    """show the build directories"""
    def _exec(self, proj, args):
        proj.show_build_dirs()


class show_builds(selectcmd):
    """show the builds and their properties"""
    def _exec(self, proj, args):
        proj.show_builds()


class show_targets(selectcmd):
    """show the targets of a single build"""
    def _exec(self, proj, args):
        proj.show_targets()


# -----------------------------------------------------------------------------
class create_proj(selectcmd):
    """create cmany.yml alongside CMakeLists.txt to hold project-settings"""
    def add_args(self, parser):
        super().add_args(parser)
        parser.add_argument('-o', '--output-file', default="cmany.yml",
                            help="""file where the project should be written.
                            Accepts relative or absolute paths. Relative paths
                            are taken from the current working directory.""")
    def _exec(self, proj, args):
        proj.create_proj()


class export_vs(selectcmd):
    """create CMakeSettings.json, a VisualStudio 2015+ compatible file
    outlining the project builds
    """
    def _exec(self, proj, args):
        proj.export_vs()


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    cmany_main(sys.argv[1:])



from collections import OrderedDict as odict
import re

from . import util
from .named_item import NamedItem
from .build_flags import BuildFlags
from .combination_rules import CombinationRules


# -----------------------------------------------------------------------------
class BuildItem(NamedItem):
    """A base class for build items."""

    @staticmethod
    def create(map_of_class_name_to_tuple_of_class_and_specs):
        items = BuildItemCollection()
        for cls_name, (cls, spec_list) in map_of_class_name_to_tuple_of_class_and_specs.items():
            if isinstance(spec_list, str):
                spec_list = util.splitesc_quoted(spec_list, ',')
            for s in spec_list:
                items.add_build_item(cls(s))
        items.resolve_references()
        return items

    def is_trivial(self):
        if self.name != self.default_str():
            return False
        if not self.flags.empty():
            return False
        return True

    @staticmethod
    def trivial_item(items):
        return len(items) == 1 and items[0].is_trivial()

    @staticmethod
    def no_flags_in_collection(items):
        for i in items:
            if not i.flags.empty():
                return False
        return True

    def __init__(self, spec):
        self.full_specs = spec
        self.flag_specs = []
        self.refs = []
        self.combination_rules = CombinationRules([])
        self._resolved_references = False
        spec = util.unquote(spec)
        spl = spec.split(':')
        if len(spl) == 1:
            name = spec
            self.flags = BuildFlags(name)
            super().__init__(name)
            return
        name = spl[0]
        rest = spl[1]
        # super().__init__(name)  # DON'T!!! will overwrite
        self.name = name
        self.flags = BuildFlags(name)
        spl = util.splitesc_quoted(rest, ' ')
        curr = ""
        for s in spl:
            if s[0] != '@':
                curr += " " + s
            else:
                self.refs.append(s[1:])
                if curr:
                    self.flag_specs.append(curr)
                    curr = ""
                self.flag_specs.append(s)
        if curr:
            self.flag_specs.append(curr)

    def resolve_references(self, item_collection):
        if self._resolved_references:
            return
        for s_ in self.flag_specs:
            s = s_.lstrip()
            if s[0] == '@':
                refname = s[1:]
                r = item_collection.lookup_build_item(refname, self.__class__)
                if self.name in r.refs:
                    msg = "circular references found in {} definitions: '{}'x'{}'"
                    raise Exception(msg.format(self.__class__.__name__, self.name, r.name))
                if not r._resolved_references:
                    r.resolve_references(item_collection)
                self.flags.append_flags(r.flags, append_to_name=False)
            else:
                import argparse
                from . import args as c4args
                parser = argparse.ArgumentParser()
                c4args.add_bundle_flags(parser)
                ss = util.splitesc_quoted(s, ' ')
                args = parser.parse_args(ss)
                tmp = BuildFlags('', **vars(args))
                self.flags.append_flags(tmp, append_to_name=False)
                cr = []
                if hasattr(args, 'combination_rules'):
                    cr = getattr(args, 'combination_rules')
                self.combination_rules = CombinationRules(cr)
        self._resolved_references = True

    @staticmethod
    def parse_args(v_):
        """parse comma-separated build item specs from the command line.

        An individual build item spec can have any of the following forms:
          * name
          * 'name:'
          * "name:"
          * 'name: <flag_specs...>'
          * "name: <flag_specs...>"
        So for example, any of these could be valid input to this function:
          * foo,bar
          * foo,'bar: -X "-a" "-b" (etc)'
          * foo,"bar: -X '-a' '-b' (etc)"
          * 'foo: -DTHIS_IS_FOO -X "-a"','bar: -X "-a" "-b" (etc)'
          * 'foo: -DTHIS_IS_FOO -X "-a"',bar
          * etc

        In some cases the shell (or argparse? or what?) removes quotes, so we
        have to deal with that too.
        """

        #util.lognotice("parse_args 0: input=____{}____".format(v_))

        # remove start and end quotes if there are any
        v = v_
        if util.is_quoted(v_):
            v = util.unquote(v_)

        # print("parse_args 1: unquoted=____{}____".format(v))

        if util.has_interior_quotes(v):
            # this is the simple case: we assume everything is duly delimited
            vli = util.splitesc_quoted(v, ',')
            # print("parse_args 2: vli=__{}__".format(vli))
        else:
            # in the absence of interior quotes, parsing is more complicated.
            # Does the string have ':'?
            if v.find(':') == -1:
                # no ':' was found; a simple split will nicely do
                vli = v.split(',')
                # print("parse_args 3.1: vli=__{}__".format(vli))
            else:
                # uh oh. we have ':' in the string, but no quotes in it. This
                # means we have to do it the hard way. There's probably a
                # less hard way, but for now this is short enough.
                # print("parse_args 3.2: parsing manually...")
                vli = []
                withc = False
                b = 0
                lastcomma = -1
                for i, c in enumerate(v):
                    if c == ',':
                        if not withc:
                            vli.append(v[b:i])
                            b = i + 1
                        # print("parse_args 3.2.1:  ','@ i={}:  v[b:i]={} vli={}".format(i, v[b:i], vli))
                        lastcomma = i
                    elif c == ':':
                        if not withc:
                            withc = True
                        else:
                            vli.append(v[b:(lastcomma + 1)])
                        b = lastcomma + 1
                        # print("parse_args 3.2.2:  ':'@ i={}:  v[b:i]={} vli={}".format(i, v[b:i], vli))
                rest = v[b:]
                if rest:
                    vli.append(rest)
                # print("parse_args 3.2.3: rest={} vli={}".format(rest, vli))

        # print("parse_args 4: vli=", vli)
        # unquote split elements
        vli = [util.unquote(v).strip(',') for v in vli]
        # util.logdone("parse_args 4: input=____{}____ output=__{}__".format(v_, vli))
        return vli

    def save_config(self, yml_node):
        if not self.flags.empty():
            self.flags.save_config(yml_node)

    def load_config(self, yml_node):
        self.flags.load_config(yml_node)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class BuildItemCollection(odict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.item_names = odict()

    def add_build_item(self, item):
        # convert the class name to snake case and append s for plural
        # eg Variant-->variants and BuildType --> build_types
        # http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
        cls = item.__class__
        cls_name = cls.__name__
        cls_name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', cls_name)
        cls_name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', cls_name)
        cls_name = cls_name.lower() + 's'
        # add the item to the list of the class name
        if not self.get(cls_name):
            self[cls_name] = []
        self[cls_name].append(item)
        # store the
        if not hasattr(self, 'collections'):
            setattr(self, 'collections', [])
        if not cls_name in self.collections:
            self.collections.append(cls_name)
        # add the object to a map to accelerate lookups by item name
        if not self.item_names.get(item.name):
            self.item_names[item.name] = []
        self.item_names[item.name].append(item)

    def lookup_build_item(self, item_name, item_cls=None):
        items = self.item_names[item_name]
        # first look for items with the same class and same name
        for i in items:
            if i.name == item_name and i.__class__ == item_cls:
                return i
        # now look for just same name
        for i in items:
            if i.name == item_name:
                return i
        # at least one must be found
        msg = "item not found: {} (class={})".format(item_name,
                                                     item_cls.__name__)
        raise Exception(msg)

    def resolve_references(self):
        for c in self.collections:
            for item in self[c]:
                item.resolve_references(self)



#!/usr/bin/env python3

import os
import glob
import json
import copy
import timeit
from collections import OrderedDict as odict

from ruamel import yaml as yaml
from ruamel.yaml.comments import CommentedMap as CommentedMap

from . import util
from . import conf

from .build_flags import BuildFlags
from .build_item import BuildItem
from .build_type import BuildType
from .system import System
from .architecture import Architecture
from .compiler import Compiler
from .variant import Variant
from .build import Build

from .combination_rules import CombinationRules
from .cmake import getcachevars


# -----------------------------------------------------------------------------
class Project:

    def __init__(self, **kwargs):

        self.kwargs = kwargs

        proj_dir = kwargs.get('proj_dir', os.getcwd())
        proj_dir = os.getcwd() if proj_dir == "." else proj_dir
        if not os.path.isabs(proj_dir):
            proj_dir = os.path.abspath(proj_dir)
        self.root_dir = proj_dir

        def _getdir(attr_name, default):
            d = kwargs.get(attr_name)
            if d is None:
                d = os.path.join(os.getcwd(), default)
            else:
                if not os.path.isabs(d):
                    d = os.path.join(os.getcwd(), d)
            return d
        self.build_dir = _getdir('build_dir', 'build')
        self.install_dir = _getdir('install_dir', 'install')

        self.cmakelists = util.chkf(self.root_dir, "CMakeLists.txt")
        self.num_jobs = kwargs.get('jobs')
        self.targets = kwargs.get('target')

        self.load_configs()

        s, a, c, t, v = __class__.get_build_items(**kwargs)

        cr = CombinationRules(kwargs.get('combination_rules', []))
        combs = cr.valid_combinations(s, a, c, t, v)
        self.combination_rules = cr

        self.builds = []
        for s_, a_, c_, t_, v_ in combs:
            self.add_build(s_, a_, c_, t_, v_)

        self.systems = s
        self.architectures = a
        self.compilers = c
        self.build_types = t
        self.variants = v

        # add new build params as needed to deal with adjusted builds
        def _addnew(b, name):
            a = getattr(b, name)
            ali = getattr(self, name + 's')
            if not [elm for elm in ali if str(elm) == str(a)]:
                ali.append(a)
        for b in self.builds:
            if not b.adjusted:
                continue
            _addnew(b, 'system')
            _addnew(b, 'architecture')
            _addnew(b, 'build_type')
            _addnew(b, 'compiler')
            _addnew(b, 'variant')

    @staticmethod
    def get_build_items(**kwargs):
        d = odict()
        for c, cls in (
                ('systems', System),
                ('architectures', Architecture),
                ('compilers', Compiler),
                ('build_types', BuildType),
                ('variants', Variant)):
            d[c] = (cls, kwargs.get(c))
        coll = BuildItem.create(d)
        s = coll['systems']
        a = coll['architectures']
        c = coll['compilers']
        t = coll['build_types']
        v = coll['variants']
        return s, a, c, t, v

    def load_configs(self):
        seq = [os.path.join(d, "cmany.yml") for d in (
            conf.CONF_DIR, conf.USER_DIR, self.root_dir)]
        if self.kwargs.get('no_default_config'):
            seq = []
        for f in self.kwargs['config_file']:
            if not os.path.isabs(f):
                f = os.path.join(self.root_dir, f)
            if not os.path.exists(f):
                raise Exception(f + ": does not exist")
            seq.append(f)
        self.configs = conf.Configs.load_seq(seq)

    def save_configs(self):
        c = Configs()

    def create_proj(self):
        yml = CommentedMap()
        yml['project'] = CommentedMap()
        #
        def _add(name):
            items = getattr(self, name)
            #if BuildItem.trivial_item(items):
            #    yml['project'][name] = "_default_"
            #elif BuildItem.no_flags_in_collection(items):
            if BuildItem.no_flags_in_collection(items):
                out = []
                for s in items:
                    out.append(s.name)
                yml['project'][name] = out
            else:
                out = []
                for s in items:
                    cm = CommentedMap()
                    cm[s.name] = CommentedMap()
                    s.save_config(cm[s.name])
                    out.append(cm)
                yml['project'][name] = out
        #
        _add('systems')
        _add('architectures')
        _add('compilers')
        _add('build_types')
        _add('variants')
        txt = yaml.round_trip_dump(yml)
        fn = self.kwargs['output_file']
        if not os.path.isabs(fn):
            fn = os.path.join(self.root_dir, fn)
        with open(fn, "w") as f:
            f.write(txt)

    def add_build(self, system, arch, compiler, build_type, variant):
        # duplicate the build items, as they may be mutated due
        # to translation of their flags for the compiler
        def _dup_item(item):
            i = copy.deepcopy(item)
            i.flags.resolve_flag_aliases(compiler, aliases=self.configs.flag_aliases)
            return i
        s = _dup_item(system)
        a = _dup_item(arch)
        t = _dup_item(build_type)
        c = _dup_item(compiler)
        v = _dup_item(variant)
        #
        f = BuildFlags('all_builds', **self.kwargs)
        f.resolve_flag_aliases(compiler, aliases=self.configs.flag_aliases)

        # create the build
        b = Build(self.root_dir, self.build_dir, self.install_dir,
                  s, a, t, c, v, f,
                  self.num_jobs, dict(self.kwargs))

        # When a build is created, its parameters may have been adjusted
        # because of an incompatible generator specification.
        # So drop this build if an equal one already exists
        if b.adjusted and self.exists(b):
            return False  # a similar build already exists

        # finally, this.
        self.builds.append(b)
        return True  # build successfully added

    def exists(self, build):
        for b in self.builds:
            if str(b.tag) == str(build.tag):
                return True
        return False

    def select(self, **kwargs):
        out = [b for b in self.builds]
        def _h(kw, attr):
            global out
            g = kwargs.get(kw)
            if g is not None:
                lo = []
                for b in out:
                    if str(getattr(b, attr)) == str(g):
                        lo.append(b)
                out = lo
        _h("sys", "system")
        _h("arch", "architecture")
        _h("compiler", "compiler")
        _h("build_type", "build_type")
        _h("variant", "variant")
        return out

    def create_tree(self, **restrict_to):
        builds = self.select(**restrict_to)
        for b in builds:
            b.create_dir()
            b.create_preload_file()
            # print(b, ":", d)

    def configure(self, **restrict_to):
        if not os.path.exists(self.build_dir):
            os.makedirs(self.build_dir)
        self._execute(Build.configure, "Configure", silent=False, **restrict_to)

    def build(self, **restrict_to):
        def do_build(build):
            build.build(self.targets)
        self._execute(do_build, "Build", silent=False, **restrict_to)

    def clean(self, **restrict_to):
        self._execute(Build.clean, "Clean", silent=False, **restrict_to)

    def install(self, **restrict_to):
        self._execute(Build.install, "Install", silent=False, **restrict_to)

    def run_cmd(self, cmd, **restrict_to):
        cmds = util.splitesc_quoted(cmd, ' ')
        def _run_cmd(b):
            with util.setcwd(b.builddir):
                util.runsyscmd(cmds)
        self._execute(_run_cmd, "Run cmd", silent=False, **restrict_to)

    def export_vs(self):
        confs = []
        for b in self.builds:
            confs.append(b.json_data())
        jd = odict([('configurations', confs)])
        with open(self.configfile, 'w') as f:
            json.dump(jd, f, indent=2)

    def show_vars(self, varlist):
        varv = odict()
        pat = os.path.join(self.build_dir, '*', 'CMakeCache.txt')
        g = glob.glob(pat)
        md = 0
        mv = 0
        for p in g:
            d = os.path.dirname(p)
            b = os.path.basename(d)
            md = max(md, len(b))
            vars = getcachevars(d, varlist)
            for k, v in vars.items():
                sk = str(k)
                if not varv.get(sk):
                    varv[sk] = odict()
                varv[sk][b] = v
                mv = max(mv, len(sk))
        #
        fmt = "{:" + str(mv) + "}[{:" + str(md) + "}]={}"
        for var, sysvalues in varv.items():
            for s, v in sysvalues.items():
                print(fmt.format(var, s, v))

    def show_build_names(self):
        for b in self.builds:
            print(b)

    def show_build_dirs(self):
        for b in self.builds:
            print(b.builddir)

    def show_builds(self):
        for b in self.builds:
            b.show_properties()

    def show_targets(self):
        for t in self.builds[0].get_targets():
            print(t)

    def _execute(self, fn, msg, silent, **restrict_to):
        builds = self.select(**restrict_to)
        durations = odict()
        num = len(builds)
        if not silent:
            if num == 0:
                print("no builds selected")
        if num == 0:
            return
        if not silent:
            if num > 1:
                util.lognotice("")
                util.lognotice("===============================================")
                util.lognotice(msg + ": start", num, "builds:")
                for b in builds:
                    util.lognotice(b)
                util.lognotice("===============================================")
        for i, b in enumerate(builds):
            if not silent:
                if i > 0:
                    util.lognotice("\n")
                util.lognotice("-----------------------------------------------")
                if num > 1:
                    util.lognotice(msg + ": build #{} of {}:".format(i+1, num), b)
                else:
                    util.lognotice(msg, b)
                util.lognotice("-----------------------------------------------")
            # this is where it happens
            t = timeit.default_timer()
            fn(b)  # <-- here
            t = timeit.default_timer() - t
            hrt = util.human_readable_time(t)
            durations[b] = (t, hrt)
            if num > 1:
                info = "finished build #{} of {} ({})".format(i + 1, num, hrt)
            else:
                info = "finished building ({})".format(hrt)
            util.logdone(msg + ": " + info + ":",  b)
        if not silent:
            util.lognotice("-----------------------------------------------")
            if num > 1:
                util.logdone(msg + ": finished", num, "builds:")
                tot = 0.
                for _, (d, _) in durations.items():
                    tot += d
                for b in builds:
                    dur, hrt = durations[b]
                    times = "({}, {:.3f}%, {:.3f}x avg)".format(
                        hrt, dur/tot*100., dur/(tot/float(num))
                    )
                    util.logdone(b, times)
                util.logdone("total time: {}".format(util.human_readable_time(tot)))
                util.lognotice("===============================================")



#!/usr/bin/env python3

import os
import re
import sys
import subprocess
import platform
import collections
import copy

import colorama #from colorama import Fore, Back, Style, init
colorama.init()

_debug_mode = False


def supports_color():
    """
    Returns True if the running system's terminal supports color, and False
    otherwise.
    Taken from http://stackoverflow.com/questions/7445658/
    """
    plat = sys.platform
    supported_platform = plat != 'Pocket PC' and (plat != 'win32' or
                                                  'ANSICON' in os.environ)
    # isatty is not always implemented, #6223.
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
    if not supported_platform or not is_a_tty:
        return False
    return True
    # also this: https://gist.github.com/ssbarnea/1316877

_suppress_colors = False
def suppress_colors():
    global _suppress_colors
    _suppress_colors = True

cmany_colored_output = (not _suppress_colors) and supports_color()


def color_log(style, *args, **kwargs):
    if cmany_colored_output:
        print(style, sep='', end='')
        print(*args, **kwargs)
        print(colorama.Style.RESET_ALL, sep='', end='')
    else:
        print(*args, **kwargs)


def logdbg(*args, **kwargs):
    if _debug_mode:
        print(*args, **kwargs)


def loginfo(*args, **kwargs):
    color_log(colorama.Fore.CYAN + colorama.Style.BRIGHT, *args, **kwargs)


def lognotice(*args, **kwargs):
    color_log(colorama.Fore.BLUE + colorama.Style.BRIGHT, *args, **kwargs)


def logdone(*args, **kwargs):
    color_log(colorama.Fore.GREEN + colorama.Style.BRIGHT, *args, **kwargs)


def logwarn(*args, **kwargs):
    color_log(colorama.Fore.YELLOW + colorama.Style.BRIGHT, *args, **kwargs)


def logerr(*args, **kwargs):
    color_log(colorama.Fore.RED + colorama.Style.BRIGHT, *args, **kwargs)


def logcmd(*args, **kwargs):
    # print(*args, **kwargs)
    print("--------")
    color_log(colorama.Fore.WHITE + colorama.Style.BRIGHT, *args, **kwargs)
    # this print here is needed to prevent the command output
    # from being colored. Need to address this somehow.
    print("--------")


# -----------------------------------------------------------------------------
def human_readable_time(seconds):
    if seconds < 60.:
        return '{:.3g}s'.format(seconds)
    elif seconds < 3600.:
        mins = int(seconds / 60.)
        secs = int(seconds - 60.*mins)
        return '{}m {}s'.format(mins, secs)
    else:
        hours = int(seconds / 3600.)
        mins = int((seconds - hours*3600.)/60.)
        secs = int(seconds - hours*3600. - mins*60.)
        return '{}:{} {}\''.format(hours, mins, secs)


# -----------------------------------------------------------------------------
def sys_str():
    if sys.platform == "linux" or sys.platform == "linux2":
        return "linux"
    elif sys.platform == "darwin":
        return "mac"
    elif sys.platform == "win32":
        return "windows"
    else:
        raise Exception("unknown system")


def in_windows():
    return sys.platform == "win32"


def in_unix():
    return sys.platform in ("linux", "linux2", "darwin")


def set_executable(filename):
    import stat
    st = os.stat(filename)
    os.chmod(filename, st.st_mode | stat.S_IEXEC)
    # if in_unix():
    #     with open(n[0], 'r') as f:
    #        mode = os.fstat(f.fileno()).st_mode
    #         mode |= 0o111
    #         os.fchmod(f.fileno(), mode & 0o7777)
    # elif in_windows():
    #     import stat
    #     st = os.stat('somefile')
    #     os.chmod('somefile', st.st_mode | stat.S_IEXEC)


def in_64bit():
    """return True if in a 64-bit architecture"""
    # http://stackoverflow.com/a/12578715/5875572
    machine = platform.machine()
    if machine.endswith('64'):
        return True
    elif machine.endswith('86'):
        return False
    raise Exception("unknown platform architecture")
    # return (struct.calcsize('P') * 8) == 64


def in_32bit():
    """return True if in a 32-bit architecture"""
    # http://stackoverflow.com/a/12578715/5875572
    machine = platform.machine()
    if machine.endswith('64'):
        return False
    elif machine.endswith('86'):
        return True
    raise Exception("unknown platform architecture")
    # return (struct.calcsize('P') * 8) == 32


def is_quoted(s):
    if s is None or len(s) == 0:
        return False
    if not (s[0] in "'\"" and s[-1] == s[0]):
        return False
    # At this point we know the string starts and ends
    # with matching quotes. But we don't know if the last quote
    # closes the first one.
    level = 0
    open = False
    for i, c in enumerate(s):
        is_escaped = i > 0 and s[i - 1] == '\\'
        if c == s[0] and not is_escaped:
            if open:
                level -= 1
                if level == 0 and i != (len(s) - 1):
                    return False
            else:
                level += 1
            open = not open
    return (level == 0)


def has_interior_quotes(s, sep=','):
    q = '[{}{}]'.format('"', "'")
    rxc = '{}{}{}'.format(q, sep, q)  # matches "," ',' ",' ',"
    rxl = '{}{}'.format(q, sep)       # matches ', ",
    rxr = '{}{}'.format(sep, q)       # matches ,' ,"
    got_em = False
    got_em = got_em or (re.search(rxc, s) is not None)
    got_em = got_em or (re.search(rxl, s) is not None)
    got_em = got_em or (re.search(rxr, s) is not None)
    return got_em


def unquote(s):
    if is_quoted(s):
        s = s[1:-1]
    return s


def splitesc_quoted(string, split_char, escape_char='\\', quote_chars='\'"'):
    """split a string at split_char, but respect (and preserve) all the
    characters inside a quote_chars pair (including escaped quote_chars and
    split_chars). split_char can also be escaped when outside of a
    quote_chars pair."""
    out = []
    i = 0
    l = len(string)
    prev = 0
    while i < l:
        is_escaped = (i > 0 and string[i - 1] == escape_char)
        c = string[i]
        # consume at once everything between quotes
        if c in quote_chars:
            j = i+1  # start counting only on the next position
            closes_quotes = False
            while j < l:
                d = string[j]
                is_escaped_j = (j > 0 and string[j - 1] == escape_char)
                if d == c and not is_escaped_j:  # found the matching quote
                    j += 1
                    closes_quotes = True
                    break
                j += 1
            # but defend against unbalanced quotes,
            # treating them as regular characters
            if not closes_quotes:
                i += 1
            else:
                s = string[prev:j]
                if s:
                    out.append(s)
                prev = j+1
                i = prev
        # when a split_char is found, append to the list
        elif c == split_char and not is_escaped:
            if i > 0 and i < l and i > prev:
                s = string[prev:i]
                if s:
                    out.append(s)
            prev = i+1
            i += 1
        # this is a regular character, just go on scanning
        else:
            i += 1
    # if there are still unread characters, append them as well
    if prev < l:
        s = string[prev:l]
        if s:
            out.append(s)
    return out


def splitesc(string, split_char, escape_char=r'\\'):
    """split a string at the given character, allowing for escaped characters
    http://stackoverflow.com/a/21107911"""
    rx = r'(?<!{}){}'.format(escape_char, split_char)
    s = re.split(rx, string)
    return s


def cslist(arg):
    """transform comma-separated arguments into a list of strings.
    commas can be escaped with backslash, \\"""
    s = splitesc(arg, ',')
    l = []
    for elm in s:
        elm = re.sub(r'\\,', r',', elm)
        l.append(elm)
    return l


def intersperse_l(delimiter, iterable):
    """put the delimiter on the left of every element of iterable"""
    it = iter(iterable)
    for x in it:
        yield delimiter
        yield x


def intersperse_r(delimiter, iterable):
    """put the delimiter on the right of every element of iterable"""
    it = iter(iterable)
    for x in it:
        yield x
        yield delimiter


def chkf(*args):
    """join the args as a path and check whether that path exists"""
    f = os.path.join(*args)
    if not os.path.exists(f):
        raise Exception("path does not exist: " + f + ". Current dir=" + os.getcwd())  # nopep8
    return f


def touch(fname, times=None):
    """change the modification and access times of a file.
    http://stackoverflow.com/questions/1158076/implement-touch-using-python
    """
    with open(fname, 'a'):
        os.utime(fname, times)


def remove_if(fname):
    """remove a file if it exists"""
    if os.path.exists(fname):
        os.remove(fname)


def which(cmd):
    """look for an executable in the current PATH environment variable"""
    if exists_and_exec(cmd):
        return cmd
    exts = ("", ".exe", ".bat") if sys.platform == "win32" else [""]
    for path in os.environ["PATH"].split(os.pathsep):
        for e in exts:
            j = os.path.join(path, cmd+e)
            if exists_and_exec(j):
                return j
    return None


def exists_and_exec(file):
    """return true if the given file exists and is executable"""
    if not os.path.exists(file):
        return False
    if not os.access(file, os.R_OK):
        return False
    if not os.access(file, os.X_OK):
        return False
    return True


def cacheattr(obj, name, function):
    """add and cache an object member which is the result of a given function.
    This is for implementing lazy getters when the function call is expensive."""
    if hasattr(obj, name):
        val = getattr(obj, name)
    else:
        val = function()
        setattr(obj, name, val)
    return val


def ctor(cls, args):
    if not isinstance(args, list):
        args = [args]
    l = []
    for i in args:
        l.append(cls(i))
    return l


# -----------------------------------------------------------------------------
def nested_lookup(dictionary, *entry):
    """get a nested entry from a dictionary"""
    try:
        if isinstance(entry, str):
            v = dictionary[entry]
        else:
            v = None
            for e in entry:
                # print("key:", e, "value:", v if v is not None else "<none yet>")
                v = v[e] if v is not None else dictionary[e]
    except:
        raise Exception("could not find entry '" +
                        "/".join(list(entry)) + "' in the dictionary")
    return v


def nested_merge(into_dct, from_dct, into_dct_is_const=True):
    """ adapted from Copied from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
    """
    out = copy.deepcopy(into_dct) if into_dct_is_const else into_dct
    for k, v in from_dct.items():
        if (k in out and isinstance(out[k], type(into_dct))
                and isinstance(from_dct[k], type(from_dct))):
            nested_merge(out[k], from_dct[k], False)
        else:
            out[k] = from_dct[k]
    return out


# -----------------------------------------------------------------------------
class setcwd:
    """temporarily change into a directory inside a with block"""

    def __init__(self, dir_, silent=True):
        self.dir = dir_
        self.silent = silent

    def __enter__(self):
        self.old = os.getcwd()
        if self.old == self.dir:
            return
        if not self.silent:
            print("Entering directory", self.dir, "(was in {})".format(self.old))
        chkf(self.dir)
        os.chdir(self.dir)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.old == self.dir:
            return
        if not self.silent:
            print("Returning to directory", self.old, "(currently in {})".format(self.dir))
        chkf(self.old)
        os.chdir(self.old)

# -----------------------------------------------------------------------------


# subprocess.run() was introduced only in Python 3.5,
# so we provide a replacement implementation to use in older Python versions.
# See http://stackoverflow.com/a/40590445
if sys.version_info >= (3, 5):
    sprun = subprocess.run
else:


    if sys.version_info < (3, 3):
        msg = 'cmany requires at least Python 3.3. Current version is {}. Sorry.'
        sys.exit(msg.format(sys.version_info))


    class CompletedProcess:
        def __init__(self, returncode, args, stdout, stderr):
            self.args = args
            self.returncode = returncode
            self.stdout = stdout
            self.stderr = stderr
        def check_returncode(self):
            if self.returncode:
                if sys.version_info >= (3, 5):
                    raise subprocess.CalledProcessError(
                        self.returncode, self.args, self.stdout, self.stderr)
                else:
                    out = ""
                    if self.stdout:
                        out += self.stdout
                    if self.stderr:
                        out += self.stderr
                    raise subprocess.CalledProcessError(
                        self.returncode, self.args, out)

    def subprocess_run_impl(*popenargs, input=None, check=False, **kwargs):
        if input is not None:
            if 'stdin' in kwargs:
                raise ValueError('stdin and input arguments may not both be used.')
            kwargs['stdin'] = subprocess.PIPE
        process = subprocess.Popen(*popenargs, **kwargs)
        try:
            stdout, stderr = process.communicate(input)
        except:
            process.kill()
            process.wait()
            raise
        retcode = process.poll()
        if check and retcode:
            raise subprocess.CalledProcessError(
                retcode, process.args, output=stdout, stderr=stderr)
        return CompletedProcess(args=process.args, returncode=retcode,
                                stdout=stdout, stderr=stderr)

    # point sprun to our implementation
    sprun = subprocess_run_impl


def runsyscmd(cmd, echo_cmd=True, echo_output=True, capture_output=False, as_bytes_string=False):
    """run a system command. Note that stderr is interspersed with stdout"""

    if not isinstance(cmd, list):
        raise Exception("the command must be a list with each argument a different element in the list")

    if echo_cmd:
        scmd = cmd
        if not isinstance(cmd, str):
            scmd = ""
            for a in cmd:
                if ' ' in a:
                    if in_windows():
                        a = '"' + a + '"'
                    else:
                        a = re.sub(r' ', r'\\ ', a)
                scmd += " " + a
        logcmd('$' + scmd)

    if as_bytes_string:
        if capture_output:
            result = sprun(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDERR)
            result.check_returncode()
            return result.stdout
        else:
            result = sprun(cmd)
            result.check_returncode()
    else:
        if not echo_output:
            result = sprun(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                           universal_newlines=True)
            result.check_returncode()
            if capture_output:
                return str(result.stdout)
        elif echo_output:
            if capture_output:
                result = sprun(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                               universal_newlines=True)
                result.check_returncode()
                return str(result.stdout)
            else:
                result = sprun(cmd, universal_newlines=True)
                result.check_returncode()



import re
import os.path
import glob

from . import cmake
from . import vsinfo

from .build_item import BuildItem


# -----------------------------------------------------------------------------
class Generator(BuildItem):

    """Visual Studio aliases example:
    vs2013: use the bitness of the current system
    vs2013_32: use 32bit version
    vs2013_64: use 64bit version
    """

    @staticmethod
    def default_str():
        """get the default generator from cmake"""
        s = cmake.CMakeSysInfo.generator()
        return s

    def __init__(self, name, build, num_jobs):
        if name.startswith('vs'):
            name = vsinfo.to_gen(name)
        self.alias = name
        super().__init__(name)
        self.num_jobs = num_jobs
        self.is_makefile = name.endswith("Makefiles")
        self.is_ninja = name.endswith("Ninja")
        self.is_msvc = name.startswith("Visual Studio")
        self.build = build
        #
        self.sysinfo_name = self.name
        if self.is_msvc:
            ts = build.compiler.vs.toolset
            self.sysinfo_name += (' ' + ts) if ts is not None else ""
        # these vars would not change cmake --system-information
        # self.full_name += " ".join(self.build.flags.cmake_vars)

    def configure_args(self, for_json=False):
        if self.name != "":
            if self.is_msvc and self.build.compiler.vs.toolset is not None:
                if for_json:
                    args = '-T ' + self.build.compiler.vs.toolset
                else:
                    args = ['-G', self.name, '-T', self.build.compiler.vs.toolset]
            else:
                if for_json:
                    args = ''
                else:
                    args = ['-G', self.name]
        else:
            if for_json:
                args = ''
            else:
                args = []
        # cmake vars are explicitly set in the preload file
        # args += self.build.flags.cmake_flags
        return args

    def cmd(self, targets):
        if self.is_makefile:
            return ['make', '-j', str(self.num_jobs)] + targets
        else:
            bt = str(self.build.build_type)
            if len(targets) > 1:
                msg = ("Building multiple targets with this generator is not "
                       "implemented. "
                       "cmake --build cannot handle multiple --target " +
                       "invokations. A generator-specific command must be "
                       "written to handle multiple targets with this "
                       "generator " + '("{}")'.format(self.name))
                raise Exception(msg)
            if not self.is_msvc:
                cmd = ['cmake', '--build', '.', '--target', targets[0], '--config', bt]
            else:
                # if a target has a . in the name, it must be substituted for _
                targets_safe = [re.sub(r'\.', r'_', t) for t in targets]
                if len(targets_safe) != 1:
                    raise Exception("msbuild can only build one target at a time: was " + str(targets_safe))
                t = targets_safe[0]
                pat = os.path.join(self.build.builddir, t + '*.vcxproj')
                projs = glob.glob(pat)
                if len(projs) == 0:
                    msg = "could not find vcx project for this target: {} (glob={}, got={})".format(t, pat, projs)
                    raise Exception(msg)
                elif len(projs) > 1:
                    msg = "multiple vcx projects for this target: {} (glob={}, got={})".format(t, pat, projs)
                    raise Exception(msg)
                proj = projs[0]
                cmd = [self.build.compiler.vs.msbuild, proj,
                       '/property:Configuration='+bt,
                       '/maxcpucount:' + str(self.num_jobs)]
            return cmd

    def install(self):
        bt = str(self.build.build_type)
        return ['cmake', '--build', '.', '--config', bt, '--target', 'install']

    """
    generators: https://cmake.org/cmake/help/v3.7/manual/cmake-generators.7.html

    Unix Makefiles
    MSYS Makefiles
    MinGW Makefiles
    NMake Makefiles
    Ninja
    Watcom WMake
    CodeBlocks - Ninja
    CodeBlocks - Unix Makefiles
    CodeBlocks - MinGW Makefiles
    CodeBlocks - NMake Makefiles
    CodeLite - Ninja
    CodeLite - Unix Makefiles
    CodeLite - MinGW Makefiles
    CodeLite - NMake Makefiles
    Eclipse CDT4 - Ninja
    Eclipse CDT4 - Unix Makefiles
    Eclipse CDT4 - MinGW Makefiles
    Eclipse CDT4 - NMake Makefiles
    KDevelop3
    KDevelop3 - Unix Makefiles
    Kate - Ninja
    Kate - Unix Makefiles
    Kate - MinGW Makefiles
    Kate - NMake Makefiles
    Sublime Text 2 - Ninja
    Sublime Text 2 - Unix Makefiles
    Sublime Text 2 - MinGW Makefiles
    Sublime Text 2 - NMake Makefiles

    Visual Studio 6
    Visual Studio 7
    Visual Studio 7 .NET 2003
    Visual Studio 8 2005 [Win64|IA64]
    Visual Studio 9 2008 [Win64|IA64]
    Visual Studio 10 2010 [Win64|IA64]
    Visual Studio 11 2012 [Win64|ARM]
    Visual Studio 12 2013 [Win64|ARM]
    Visual Studio 14 2015 [Win64|ARM]
    Visual Studio 15 2017 [Win64|ARM]

    Green Hills MULTI
    Xcode
    """



import os
import re
import tempfile

from .build_item import BuildItem
from .system import System
from .cmake import CMakeSysInfo
from . import util
from . import vsinfo


# -----------------------------------------------------------------------------
class Compiler(BuildItem):
    """Represents a compiler choice"""

    @staticmethod
    def default():
        return Compiler(__class__.default_str())

    @staticmethod
    def default_str():
        if str(System.default()) != "windows":
            cpp = CMakeSysInfo.cxx_compiler()
        else:
            vs = vsinfo.find_any()
            cpp = vs.name if vs is not None else CMakeSysInfo.cxx_compiler()
        return cpp

    def is_trivial(self):
        """reimplement BuildItem.is_trivial because our name is altered from the
        path to compiler_name+version"""
        if self.path != self.default_str():
            return False
        if not self.flags.empty():
            return False
        return True

    def __init__(self, spec):
        if util.is_quoted(spec):
            spec = util.unquote(spec)
        spl = spec.split(':')
        path = spl[0]
        if path.startswith("vs") or path.startswith("Visual Studio"):
            vs = vsinfo.VisualStudioInfo(path)
            self.vs = vs
            path = vs.cxx_compiler
        else:
            # in windows, defend against paths written for example as
            # C:\path\to\compiler. The split is inappropriate here.
            if (util.in_windows() and len(path) == 1 and len(spl) > 1 and
                (spl[1][0] == '/' or spl[1][0] == '\\')):
                path = spec
                spl = [path]
            p = util.which(path)
            if p is None:
                raise Exception("compiler not found: " + path)
            # if p != path:
            #     print("compiler: selected {} for {}".format(p, path))
            path = os.path.abspath(p)
        name, version, version_full = self.get_version(path)
        self.shortname = name
        self.gcclike = self.shortname in ('gcc', 'clang', 'icc', 'g++', 'clang++', 'icpc')
        self.is_msvc = self.shortname.startswith('vs')
        if not self.is_msvc:
            name += version
        self.path = path
        self.version = version
        self.version_full = version_full
        self.name_for_flags = self.shortname
        if self.is_msvc:
            if self.vs.is_clang:
                self.name_for_flags = 'clang'
            else:
                self.name_for_flags = 'vs'
        # don't forget: a build item should be initialized with the full spec
        if len(spl) > 1:
            # change the spec to reflect the real compiler name
            spec = name + ": " + ":".join(spl[1:])
        else:
            spec = name
        super().__init__(spec)
        # maybe we should use a different way to get the c compiler.
        # For now this will do.
        self.c_compiler = __class__.get_c_compiler(self.shortname, self.path)

    @staticmethod
    def get_c_compiler(shortname, cxx_compiler):
        # if cxx_compiler.endswith("c++") or cxx_compiler.endswith('c++.exe'):
        #     cc = re.sub(r'c\+\+', r'cc', cxx_compiler)
        if shortname.startswith('vs') or re.search(r'sual Studio', cxx_compiler):
            cc = cxx_compiler
        elif shortname == "icc" or shortname == "icpc":
            cc = re.sub(r'icpc', r'icc', cxx_compiler)
        elif shortname == "gcc" or shortname == "g++":
            if re.search(r'g\+\+', cxx_compiler):
                cc = re.sub(r'g\+\+', r'gcc', cxx_compiler)
            else:
                cc = re.sub(r'c\+\+', r'cc', cxx_compiler)
        elif shortname == "clang" or shortname == "clang++":
            if re.search(r'clang\+\+', cxx_compiler):
                cc = re.sub(r'clang\+\+', r'clang', cxx_compiler)
            else:
                cc = re.sub(r'c\+\+', r'cc', cxx_compiler)
        elif shortname.startswith("arm-"):
            cc = re.sub(r'g\+\+', 'gcc', cxx_compiler)
        elif re.search(r"c\+\+", shortname) or re.search(r'apple_llvm', shortname):
            cc = re.sub(r"c\+\+", "cc", cxx_compiler)
        else:
            cc = "cc"
        return cc

    def get_version(self, path):
        # a function to silently run a system command
        def slntout(cmd):
            out = util.runsyscmd(cmd, echo_cmd=False,
                                 echo_output=False, capture_output=True)
            out = out.strip("\n")
            return out
        # is this visual studio?
        if hasattr(self, "vs"):
            return self.vs.name, str(self.vs.year), self.vs.name
        # other compilers
        # print("cmp: found compiler:", path)
        out = slntout([path, '--version'])
        version_full = out.split("\n")[0]
        splits = version_full.split(" ")
        name = splits[0].lower()
        # print("cmp: version:", name, "---", version_full, "---")
        vregex = r'(\d+\.\d+)\.\d+'
        base = os.path.basename(path)
        # print("cmp base:", base)
        if base.startswith("c++") or base.startswith("cc"):
            try:  # if this fails, just go on. It's not really needed.
                with tempfile.NamedTemporaryFile(suffix=".cpp", prefix="cmany.", delete=False) as f:
                    macros = slntout([path, '-dM', '-E', f.name])
                    os.unlink(f.name)
                macros = macros.split("\n")
            except:
                macros = []
            for m in sorted(macros):
                if re.search("#define __clang__", m):
                    name = "clang++" if re.search(r"\+\+", path) else "clang"
                    break
                elif re.search("#define __GNUC__", m):
                    name = "g++" if re.search(r"\+\+", path) else "gcc"
                    break
        if name.startswith("g++") or name.startswith("gcc"):
            # print("g++: version:", name, name.find('++'))
            name = "g++" if name.find('++') != -1 else 'gcc'
            # print("g++: version:", name, name.find('++'))
            version = slntout([path, '-dumpversion'])
            version = re.sub(vregex, r'\1', version)
            # print("gcc version:", version, "---")
        elif name.startswith("clang"):
            name = "clang++" if path.find('clang++') != -1 else 'clang'
            if re.search('Apple LLVM', version_full):
                name = "apple_llvm"
                version = re.sub(r'Apple LLVM version ' + vregex + '.*', r'\1', version_full)
                print("apple_llvm version:", version, "---")
            else:
                version = re.sub(r'clang version ' + vregex + '.*', r'\1', version_full)
                # print("clang version:", version, "---")
        elif name.startswith("icpc") or name.startswith("icc"):
            name = "icc" if name.startswith("icc") else "icpc"
            if re.search(r'icpc \(ICC\) ' + vregex + '.*', version_full):
                version = re.sub(r'icpc \(ICC\) ' + vregex + '.*', r'\1', version_full)
            else:
                version = re.sub(r'icc \(ICC\) ' + vregex + '.*', r'\1', version_full)
            # print("icc version:", version, "---")
        else:
            version = slntout([path, '-dumpversion'])
            version = re.sub(vregex, r'\1', version)
        #
        return name, version, version_full

    def make_32bit(self):
        if self.gcclike:
            self.flags.cflags.append('-m32')
            self.flags.cxxflags.append('-m32')

    def make_64bit(self):
        if self.gcclike:
            self.flags.cflags.append('-m64')
            self.flags.cxxflags.append('-m64')



#!/usr/bin/env python3

import unittest as ut
import subtest_fix
import tempfile
import os
import argparse
from collections import OrderedDict as odict

from c4.cmany import conf,  flags, util, args as c4args


# -----------------------------------------------------------------------------
class Test00FlagState(ut.TestCase):

    def test00CompilersShouldBeListedInFlags(self):
        for tcn, (comps, cflags, yml) in test_cases.items():
            with self.subTest(test_case=tcn):
                for k, v in cflags.items():
                    for c in comps:
                        self.assertTrue(c in v.compilers, str(c) + " is not listed in " + str(v.compilers))


# -----------------------------------------------------------------------------
class Test01Dump(ut.TestCase):

    def test(self):
        for tcn, (comps, cflags, yml) in test_cases.items():
            with self.subTest(test_case=tcn):
                txt = flags.dump_yml(comps, cflags)
                self.assertEqual(yml, txt, tcn)


# -----------------------------------------------------------------------------
class Test02Load(ut.TestCase):

    def test00RunTestCases(self):
        for tcn, (rcomps, rflags, yml) in test_cases.items():
            with self.subTest(test_case=tcn):
                (comps, cflags) = flags.load_txt(yml)
                #print()
                #print(tcn, rcomps, comps)
                #print(tcn, rflags, cflags)
                #for (k1, v1), (k2, v2) in zip(rflags.items(), cflags.items()):
                #    print(tcn, "k1v1", k1, v1)
                #    print(tcn, "k2v2", k2, v2)
                #    for c1, c2 in zip(v1.compilers, v2.compilers):
                #        print(tcn, "comp", c1, c2)
                self.assertEqual(comps, rcomps, tcn)
                self.assertTrue(same_flags(cflags, rflags), tcn)

    def test01DoCommasWork(self):
        yml = """c++11:
    gcc,clang,icc: -std=c++11
    vs: ''
"""
        c, f = flags.load_txt(yml)
        self.assertTrue('c++11' in f)
        f11 = f['c++11']
        self.assertTrue(hasattr(f11, 'gcc'))
        self.assertTrue(hasattr(f11, 'icc'))
        self.assertTrue(hasattr(f11, 'clang'))
        self.assertEqual(f11.gcc, '-std=c++11')
        self.assertEqual(f11.icc, '-std=c++11')
        self.assertEqual(f11.clang, '-std=c++11')


# -----------------------------------------------------------------------------
class Test03Merge(ut.TestCase):

    def test(self):
        self._run('gcc-g3', 'clang-g3', 'gcc, clang-g3', 'clang, gcc-g3')

    def _run(self, tc1, tc2, r1into2, r2into1):
        for v1, v2, ref in ((tc1, tc2, r1into2), (tc2, tc1, r2into1)):
            tcn = "merge " + v1 + " into " + v2 + ": should be same as " + ref
            with self.subTest(test_case=tcn):
                comps1, cflags1, yml1 = test_cases[v1]
                comps2, cflags2, yml2 = test_cases[v2]
                compsr, cflagsr, ymlr = test_cases[ref]
                cflagsv = flags.merge(cflags2, cflags1)
                compsv = flags.get_all_compilers(cflagsv)
                ymlv = flags.dump_yml(compsv, cflagsv)
                self.assertTrue(same_elements_in_list(compsv, compsr), tcn)
                self.assertEqual(compsv, compsr, tcn)
                self.assertEqual(ymlv, ymlr, tcn)


# -----------------------------------------------------------------------------
class Test04FlagsIO(ut.TestCase):

    def setUp(self):
        with open(os.path.join(conf.CONF_DIR, "cmany.yml"), "r") as f:
            self.comps, self.flags = flags.load_txt(f.read())

    @staticmethod
    def _do_save(comps_, flags_, filename):
        with open(filename, 'w') as f:
            yml = flags.dump_yml(comps_, flags_)
            f.write(yml)

    @staticmethod
    def _do_load(filename):
        with open(filename, 'r') as f:
            yml = f.read()
            comps_, flags_ = flags.load_txt(yml)
        return comps_, flags_

    def test00SavedIsSameAsOriginal(self):
        for tcn, (rcomps, rflags, yml) in test_cases.items():
            with self.subTest(test_case=tcn):
                fh_, fn = tmpfile()
                fh = os.fdopen(fh_)
                __class__._do_save(rcomps, rflags, fn)
                c, f = __class__._do_load(fn)
                self.assertEqual(c, rcomps)
                self.assertEqual(len(f), len(rflags))
                self.assertEqual(list(f.keys()), list(rflags.keys()))
                for (rname, rf), (vname, vf) in zip(rflags.items(), f.items()):
                    self.assertEqual(rcomps, vf.compilers)
                    for kc in self.comps:
                        #print(rname, c, rf.get(c), vf.get(c))
                        self.assertEqual(rf.get(kc), vf.get(kc))
                fh.close()
                os.remove(fn)
                del fh

    def test01WriteReadWriteIsSame(self):
        for tcn, (rcomps, rflags, yml) in test_cases.items():
            with self.subTest(test_case=tcn):
                frefh, fn = tmpfile()
                fvalh, fn_out = tmpfile()
                fref = os.fdopen(frefh)
                fval = os.fdopen(fvalh)
                __class__._do_save(rcomps, rflags, fn)
                c, f = __class__._do_load(fn)
                __class__._do_save(c, f, fn_out)
                ref = list(fref.readlines())
                val = list(fval.readlines())
                lr = len(ref)
                lv = len(val)
                self.assertEqual(lr, lv)
                for i in range(0, max(lr, lv)):
                    if i < lr and i < lv:
                        self.assertEqual(ref[i], val[i])
                    else:
                        break
                del fref, frefh
                del fval, fvalh
                os.remove(fn)
                os.remove(fn_out)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test10Flags(ut.TestCase):

    def test01_cmake_vars(self):
        self._do_separate_test('-V', '--cmake-vars')

    def test02_defines(self):
        self._do_separate_test('-D', '--defines')

    def test03_cxxflags(self):
        self._do_separate_test('-X', '--cxxflags')

    def test04_cflags(self):
        self._do_separate_test('-C', '--cflags')

    def _do_separate_test(self, shortopt, longopt):
        n = longopt[2:].replace('-', '_')
        for o in (shortopt, longopt):
            def _c(a, r): self.check_one(n, o, a, r)

            _c('{} VAR1', ['VAR1'])
            _c('{} VAR2,VAR3', ['VAR2', 'VAR3'])
            _c('{} VAR4 {} VAR5', ['VAR4', 'VAR5'])
            _c('{} VAR6,VAR7 {} VAR8,VAR9', ['VAR6', 'VAR7', 'VAR8', 'VAR9'])
            _c('{} VAR\,1', ['VAR,1'])
            _c('{} VAR\,2,VAR\,3', ['VAR,2', 'VAR,3'])
            _c('{} VAR\,4 {} VAR\,5', ['VAR,4', 'VAR,5'])
            _c('{} VAR\,6,VAR\,7 {} VAR\,8,VAR\,9', ['VAR,6', 'VAR,7', 'VAR,8', 'VAR,9'])
            _c('{} VAR1=1', ['VAR1=1'])
            _c('{} VAR2=2,VAR3=3', ['VAR2=2', 'VAR3=3'])
            _c('{} VAR4=4 {} VAR5=5', ['VAR4=4', 'VAR5=5'])
            _c('{} VAR6=6,VAR7=7 {} VAR8=8,VAR9=9', ['VAR6=6', 'VAR7=7', 'VAR8=8', 'VAR9=9'])
            _c('{} VAR1=1\,a', ['VAR1=1,a'])
            _c('{} VAR2=2\,a,VAR3=3\,a', ['VAR2=2,a', 'VAR3=3,a'])
            _c('{} VAR4=4\,a {} VAR5=5\,a', ['VAR4=4,a', 'VAR5=5,a'])
            _c('{} VAR6=6\,a,VAR7=7\,a {} VAR8=8\,a,VAR9=9\,a',
               ['VAR6=6,a', 'VAR7=7,a', 'VAR8=8,a', 'VAR9=9,a'])
            _c(['{}', 'VAR1="1 with spaces"'], ['VAR1="1 with spaces"'])
            _c(['{}', 'VAR2="2 with spaces",VAR3="3 with spaces"'],
               ['VAR2="2 with spaces"', 'VAR3="3 with spaces"'])
            _c(['{}', 'VAR4="4 with spaces"', '{}', 'VAR5="5 with spaces"'],
               ['VAR4="4 with spaces"', 'VAR5="5 with spaces"'])
            _c(['{}', 'VAR6="6 with spaces",VAR7="7 with spaces"', '{}',
                'VAR8="8 with spaces",VAR9="9 with spaces"'],
               ['VAR6="6 with spaces"', 'VAR7="7 with spaces"', 'VAR8="8 with spaces"',
                'VAR9="9 with spaces"'])
            _c(['{}', 'VAR1="1\,a with spaces"'], ['VAR1="1,a with spaces"'])
            _c(['{}', 'VAR2="2\,a with spaces",VAR3="3\,a with spaces"'],
               ['VAR2="2,a with spaces"', 'VAR3="3,a with spaces"'])
            _c(['{}', 'VAR4="4\,a with spaces"', '{}', 'VAR5="5\,a with spaces"'],
               ['VAR4="4,a with spaces"', 'VAR5="5,a with spaces"'])
            _c(['{}', 'VAR6="6\,a with spaces",VAR7="7\,a with spaces"', '{}',
                'VAR8="8\,a with spaces",VAR9="9\,a with spaces"'],
               ['VAR6="6,a with spaces"', 'VAR7="7,a with spaces"', 'VAR8="8,a with spaces"',
                'VAR9="9,a with spaces"'])
            _c('{} "-fPIC,-Wall,-O3,-Os"', ['-fPIC', '-Wall', '-O3', '-Os'])
            _c("{} '-fPIC,-Wall,-O3,-Os'", ['-fPIC', '-Wall', '-O3', '-Os'])
            _c('{} "-fPIC","-Wall","-O3","-Os"', ['-fPIC', '-Wall', '-O3', '-Os'])
            _c("{} '-fPIC','-Wall','-O3','-Os'", ['-fPIC', '-Wall', '-O3', '-Os'])

    def check_one(self, name, opt, args, ref):
        if isinstance(args, str):
            args = args.split(' ')
        args_ = args
        args = []
        for a in args_:
            args.append(a.format(opt))
        p = argparse.ArgumentParser()
        c4args.add_bundle_flags(p)
        out = p.parse_args(args)
        # print(out, kwargs)
        a = getattr(out, name)
        self.assertEqual(ref, a)
        del out

    def check_many(self, args, **ref):
        if isinstance(args, str):
            args = util.splitesc_quoted(args, ' ')
        p = argparse.ArgumentParser()
        c4args.add_bundle_flags(p)
        out = p.parse_args(args)
        # print(out, kwargs)
        for k, refval in ref.items():
            result = getattr(out, k)
            self.assertEqual(result, refval)
        del out

    def test10_mixed0(self):
        self.check_many('-X "-fPIC" -D VARIANT1', cmake_vars=[], cxxflags=['-fPIC'], cflags=[],
                        defines=['VARIANT1'])
        self.check_many('-X "-Wall" -D VARIANT2', cmake_vars=[], cxxflags=['-Wall'], cflags=[],
                        defines=['VARIANT2'])
        self.check_many('-X nortti,c++14 -D VARIANT3', cmake_vars=[], cxxflags=['nortti', 'c++14'], cflags=[],
                        defines=['VARIANT3'])
        self.check_many('-X "-fPIC,-Wl\,-rpath" -D VARIANT1', cmake_vars=[], cxxflags=['-fPIC', '-Wl,-rpath'],
                        cflags=[], defines=['VARIANT1'])

    def test11_mixed1(self):
        self.check_many('-X "-fPIC" -D VARIANT1,VARIANT_TYPE=1', cmake_vars=[], cxxflags=['-fPIC'], cflags=[],
                        defines=['VARIANT1', 'VARIANT_TYPE=1'])
        self.check_many('-X "-Wall" -D VARIANT2,VARIANT_TYPE=2', cmake_vars=[], cxxflags=['-Wall'], cflags=[],
                        defines=['VARIANT2', 'VARIANT_TYPE=2'])
        self.check_many('-X nortti,c++14 -D VARIANT3,VARIANT_TYPE=3', cmake_vars=[], cxxflags=['nortti', 'c++14'],
                        cflags=[], defines=['VARIANT3', 'VARIANT_TYPE=3'])

    def test12_mixed2(self):
        self.check_many('-X "-fPIC" -D VARIANT1,"VARIANT_TYPE=1"', cmake_vars=[], cxxflags=['-fPIC'], cflags=[],
                        defines=['VARIANT1', 'VARIANT_TYPE=1'])
        self.check_many('-X "-Wall" -D VARIANT2,"VARIANT_TYPE=2"', cmake_vars=[], cxxflags=['-Wall'], cflags=[],
                        defines=['VARIANT2', 'VARIANT_TYPE=2'])
        self.check_many('-X nortti,c++14 -D VARIANT3,"VARIANT_TYPE=3"', cmake_vars=[],
                        cxxflags=['nortti', 'c++14'], cflags=[], defines=['VARIANT3', 'VARIANT_TYPE=3'])

    def test13_mixed3(self):
        self.check_many('-X "-fPIC" -D "VARIANT1,VARIANT_TYPE=1"', cmake_vars=[], cxxflags=['-fPIC'], cflags=[],
                        defines=['VARIANT1', 'VARIANT_TYPE=1'])
        self.check_many('-X "-Wall" -D "VARIANT2,VARIANT_TYPE=2"', cmake_vars=[], cxxflags=['-Wall'], cflags=[],
                        defines=['VARIANT2', 'VARIANT_TYPE=2'])
        self.check_many('-X nortti,c++14 -D "VARIANT3,VARIANT_TYPE=3"', cmake_vars=[],
                        cxxflags=['nortti', 'c++14'], cflags=[], defines=['VARIANT3', 'VARIANT_TYPE=3'])


# -----------------------------------------------------------------------------
def tmpfile():
    return tempfile.mkstemp('.yml', 'cmany_flags-')


def same_elements_in_list(l1, l2):
    for k1, k2 in zip(l1, l2):
        if not (k1 in l2):
            return False
        if not (k2 in l1):
            return False
    return True


def same_flags(f1, f2):
    if len(f1) != len(f2):
        return False
    if not same_elements_in_list(f1.keys(), f2.keys()):
        return False
    for (f1n, f1v), (f2n, f2v) in zip(f1.items(), f2.items()):
        if not same_elements_in_list(f1v.compilers, f2v.compilers):
            return False
        for c in f1v.compilers:
            if f1v.get(c) != f2v.get(c):
                return False
    return True


# -----------------------------------------------------------------------------

def tc(name, comps, flags, **kwargs):
    yml = kwargs['yml']
    d_ = (name, (comps, flags, yml))
    return d_


def d(*args):
    l = list(args)
    return odict(l)


def f(name, *args, **kwargs):
    return name, flags.CFlag(name, *args, **kwargs)


with open(os.path.join(conf.CONF_DIR, "cmany.yml")) as fconf:
    kc, kf = flags.load_txt(fconf.read())
kyml = flags.dump_yml(kc, kf)


test_cases = d(

    tc('gcc-g', ['gcc'], d(f('g', gcc='-g')), yml="""\
g:
    gcc: -g
"""),

    tc('gcc-g3', ['gcc'], d(f('g3', gcc='-g3')), yml="""\
g3:
    gcc: -g3
"""),

    tc('clang-g3', ['clang'], d(f('g3', clang='-g3')), yml="""\
g3:
    clang: -g3
"""),

    tc('gcc, clang-g3', ['gcc', 'clang'], d(f('g3', gcc='-g3', clang='-g3')), yml="""\
g3:
    gcc,clang: -g3
"""),

    tc('clang, gcc-g3', ['clang', 'gcc'], d(f('g3', clang='-g3', gcc='-g3')), yml="""\
g3:
    clang,gcc: -g3
"""),

    tc('known_flags', kc, kf, yml=kyml)
)



from ruamel import yaml
from ruamel.yaml.comments import CommentedMap as CommentedMap

import os.path as osp
from collections import OrderedDict as odict
from . import util

CONF_DIR = osp.join(osp.dirname(__file__), "conf")
# maybe cmany is not installed. Then it must be a source distribution.
if not osp.exists(CONF_DIR):
    CONF_DIR = osp.abspath(osp.join(osp.dirname(__file__), "../../../conf"))
    assert osp.exists(CONF_DIR)

USER_DIR = osp.expanduser("~/.cmany/")

if yaml.version_info < (0, 15):
    raise Exception("cmany now requires ruamel.yaml>=0.15.0")


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Configs:

    @staticmethod
    def load_seq(file_seq):
        curr = None
        for fn in file_seq:
            if not osp.exists(fn):
                continue
            if curr is None:
                curr = Configs()
                curr.load(fn)
            else:
                tmp = Configs()
                tmp.load(fn)
                curr.merge_from(tmp)
        return curr

    def __init__(self):
        self._load_yml("""\
project: {}
config: {}
flag_aliases: {}
""")

    def _load_yml(self, yml):
        YAML = yaml.YAML()
        dump = YAML.load(yml)
        dump = odict(dump)
        for i in ('project', 'config', 'flag_aliases'):
            if dump.get(i) is None:
                dump[i] = CommentedMap()
            setattr(self, i, dump[i])
        self._dump = dump
        from . import flags as c4flags
        self.flag_aliases = c4flags.FlagAliases(yml=dump.get('flag_aliases', CommentedMap()))

    def load(self, file=None, text=None):
        """file or text"""
        if file == text:
            raise Exception("either file or text")
        if file:
            with open(file, "r") as f:
                text = f.read()
        self._load_yml(text)

    def save(self, filename):
        raise Exception("not implemented")

    @staticmethod
    def _merge(dict_recv, dict_send):
        #"""http://stackoverflow.com/questions/38987/how-to-merge-two-python-dictionaries-in-a-single-expression"""
        #z = dict_recv.copy()
        #z.update(dict_send)
        z = util.nested_merge(dict_recv, dict_send)
        return z

    def merge_from(self, other):
        for k in (list(self._dump.keys()) + list(other._dump.keys())):
            dst = self._dump.get(k)
            src = other._dump.get(k)
            def is_dict(d):
                return isinstance(d, dict) or isinstance(d, odict) or isinstance(d, CommentedMap)
            if (dst is not None and src is not None):
                if is_dict(dst) and is_dict(src):
                    self._dump[k] = __class__._merge(dst, src)
                elif is_dict(dst) or is_dict(src):
                    raise Exception("cannot merge when only one is a dict")
                else:
                    self._dump[k] = other._dump[k]
            elif src is not None:
                self._dump[k] = other._dump[k]
        self.flag_aliases = self.flag_aliases.merge_from(other.flag_aliases)

    def get_val(self, name_sub, where=None):
        if where is None:
            where = self._dump
        elms = name_sub.split('.')
        curr = where
        for e in elms:
            if curr.get(e) is not None:
                curr = curr.get(e)
            else:
                return None
        return curr

    def set_val(self, name_sub, value, where=None):
        spl = name_sub.split(".")
        child = spl[-1]
        seq = spl[:-1]
        parent_name = ".".join(seq)
        parent = self.get_val(parent_name, where)
        if parent is None:
            curr = self._dump
            for c in seq:
                ch = curr.get(c)
                if ch is None:
                    curr[c] = CommentedMap()
                    curr = curr[c]
            parent = curr
        #
        if isinstance(value, str) or isinstance(value, int) or isinstance(value, float) or isinstance(value, list):
            parent[child] = value
        elif isinstance(value, dict) or isinstance(value, odict):
            parent[child] = CommentedMap(value)

    def append_val(self, name_sub, value, where=None):
        oldval = self.get_val(name_sub, where)
        if not isinstance(oldval, list):
            raise Exception("WTF?")
        value = oldval + value
        self.set_val(name_sub, value, where)
